from __future__ import annotations

from typing import TypeVar

import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as kbackend

from innvestigate.backend.types import Layer, OptionalList, ShapeTuple, Tensor

__all__ = [
    "to_list",
    "unpack_singleton",
    "shape",
    "batch_size",
    "gradients",
    "cast_to_floatx",
    "is_not_finite",
    "safe_divide",
    "count_non_zero",
    "add_gaussian_noise",
    "apply_mask",
    "apply",
    "broadcast_np_tensors_to_keras_tensors",
    "extract_conv2d_patches",
]

_EPS = kbackend.epsilon()

T = TypeVar("T")  # Generic type, can be anything


def disable_eager_execution() -> None:
    tf.compat.v1.disable_eager_execution()


def to_list(X: OptionalList[T]) -> list[T]:
    """Wraps tensor `X` into a list, if it isn't a list of Tensors yet."""
    if isinstance(X, list):
        return X
    return [X]


def unpack_singleton(x: OptionalList[T]) -> OptionalList[T]:
    """Gets the first element of a list if it has only one value.

    Otherwise return the list.

    # Argument
        x: A list or singleton.

    # Returns
        The same list or the first element.
    """
    if isinstance(x, list) and len(x) == 1:
        return x[0]
    return x


def shape(X: Tensor) -> list[int | None]:
    """Return shape of Tensor as list of ints."""
    shape: list[int | None] = X.get_shape().as_list()
    return shape


def batch_size(X: Tensor) -> int:
    """Return batch size of Tensor as integer."""
    bs: int | None = X.get_shape()[0]
    if isinstance(bs, int):
        return bs
    raise ValueError(f"Found non-integer batch_size {bs} for Tensor {X}")


def gradients(
    Xs: OptionalList[Tensor], Ys: OptionalList[Tensor], known_Ys: OptionalList[Tensor]
) -> list[Tensor]:
    if len(Ys) != len(known_Ys):
        raise ValueError(
            "Gradient computation failesd, Ys and known_Ys not of same length"
        )

    grad = tf.gradients(Ys, Xs, grad_ys=known_Ys, stop_gradients=Xs)
    if grad is None:
        raise TypeError("Gradient computation failed, returned None.")
    return to_list(grad)


def is_not_finite(X: Tensor) -> Tensor:  # returns Tensor of dtype bool
    """Checks if tensor x is finite, if not throws an exception."""
    # x = tensorflow.check_numerics(x, "innvestigate - is_finite check")
    return tf.logical_not(tf.math.is_finite(X))


def cast_to_floatx(X: Tensor) -> Tensor:
    return tf.cast(X, kbackend.floatx())


def safe_divide(A: Tensor, B: Tensor, factor: float = _EPS) -> Tensor:
    """Divide A by B, replacing all zeroes in B with `factor`."""
    is_zero = cast_to_floatx(kbackend.equal(B, kbackend.constant(0)))
    return A / (B + factor * is_zero)


def count_non_zero(X: Tensor, axis, keepdims: bool = False) -> Tensor:
    "Count non-zero elements in tensor."
    non_zeros = cast_to_floatx(kbackend.not_equal(X, kbackend.constant(0)))
    return kbackend.sum(non_zeros, axis=axis, keepdims=keepdims)


def add_gaussian_noise(X: Tensor, mean: float = 0.0, stddev: float = 1.0) -> Tensor:
    """Add Gaussian noise to tensor `X`."""
    return X + kbackend.random_normal(
        shape=tf.shape(X), mean=mean, stddev=stddev, dtype=X.dtype
    )


def apply_mask(Xs: list[T], mask: list[bool]) -> list[T]:
    """Apply mask to list `Xs`, keeping only the elements for which
    mask is True.

    :param Xs: List to be masked.
    :type Xs: List[T]
    :param mask: Mask applied to `Xs`. If True, corresponding entry in `Xs` is kept.
    :type mask: List[bool]
    :return: Masked list.
    :rtype: List[T]
    """
    if len(Xs) != len(mask):
        raise ValueError("mask not of same length as list that is to be masked.")
    return [x for x, keep in zip(Xs, mask) if keep]


###############################################################################


def apply(layer: Layer, inputs: OptionalList[Tensor]) -> list[Tensor]:
    """
    Apply a layer to input[s].

    A flexible apply that tries to fit input to layers expected input.
    This is useful when one doesn't know if a layer expects a single tensor
    or many.

    :param layer: A Keras layer instance.
    :type layer: Layer
    :param inputs: A list of input tensors or a single tensor.
    :type inputs: OptionalList[Tensor]
    :return: Output from applying the layer to the input.
    :rtype: List[Tensor]
    """

    if isinstance(inputs, list) and len(inputs) > 1:
        try:
            ret = layer(inputs)
        except (TypeError, AttributeError) as err:
            # layer expects a single tensor.
            if len(inputs) != 1:
                raise ValueError("Layer expects only a single input!") from err
            ret = layer(inputs[0])
    else:
        ret = layer(inputs[0])

    return to_list(ret)


def broadcast_np_tensors_to_keras_tensors(
    np_tensors: float | np.ndarray | list[np.ndarray],
    keras_tensors: OptionalList[Tensor],
) -> list[np.ndarray]:
    """Broadcasts numpy tensors to the shape of Keras tensors.

    :param np_tensors: Numpy tensors that should be broadcasted.
    :type np_tensors: Union[np.ndarray, List[np.ndarray]]
    :param keras_tensors: The Keras tensors with the target shapes.
    :type keras_tensors: OptionalList[Tensor]
    :return: The broadcasted Numpy tensors.
    :rtype: List[np.ndarray]
    """

    def none_to_one(shape: ShapeTuple):
        return [1 if dim is None else dim for dim in shape]

    keras_tensors = to_list(keras_tensors)

    if isinstance(np_tensors, list):
        return [
            np.broadcast_to(n, none_to_one(kbackend.int_shape(k)))
            for k, n in zip(keras_tensors, np_tensors)
        ]
    return [
        np.broadcast_to(np_tensors, none_to_one(kbackend.int_shape(k)))
        for k in keras_tensors
    ]


def extract_conv2d_patches(X: Tensor, kernel_shape, strides, rates, padding) -> Tensor:
    """Extracts conv2d patches like TF function extract_image_patches.

    :param x: Input image.
    :param kernel_shape: Shape of the Keras conv2d kernel.
    :param strides: Strides of the Keras conv2d layer.
    :param rates: Dilation rates of the Keras conv2d layer.
    :param padding: Paddings of the Keras conv2d layer.
    :return: The extracted patches.
    """
    if kbackend.image_data_format() == "channels_first":
        X = kbackend.permute_dimensions(X, (0, 2, 3, 1))
    kernel_shape = [1, kernel_shape[0], kernel_shape[1], 1]
    strides = [1, strides[0], strides[1], 1]
    rates = [1, rates[0], rates[1], 1]
    ret = tf.compat.v1.extract_image_patches(
        X, kernel_shape, strides, rates, padding.upper()
    )

    if kbackend.image_data_format() == "channels_first":
        # TODO: check if we need to permute again.xs
        pass
    return ret



"""Check Keras Layers for properties,
e.g. if it is an input or a pooling layer"""

from __future__ import annotations

import tensorflow.keras.layers as klayers
from tensorflow import Module, keras

import innvestigate.backend as ibackend
from innvestigate.backend.types import Layer

__all__ = [
    "get_activation_search_safe_layers",
    "contains_activation",
    "contains_kernel",
    "only_relu_activation",
    "is_module",
    "is_convnet_layer",
    "is_average_pooling",
    "is_max_pooling",
    "is_input_layer",
    "is_batch_normalization_layer",
    "is_embedding_layer",
]


def get_activation_search_safe_layers():
    """
    Returns a list of keras layer that we can walk along
    in an activation search.
    """

    # Inside function to not break import if Keras changes.
    activation_search_safe_layers = (
        klayers.ELU,
        klayers.LeakyReLU,
        klayers.PReLU,
        klayers.Softmax,
        klayers.ThresholdedReLU,
        klayers.Activation,
        klayers.ActivityRegularization,
        klayers.Dropout,
        klayers.Flatten,
        klayers.Reshape,
        klayers.Add,
        klayers.GaussianNoise,
        klayers.BatchNormalization,
    )
    return activation_search_safe_layers


###############################################################################


def contains_activation(layer: Layer, activation: str = None) -> bool:
    """Check whether the layer contains an activation function of type `activation`.
    If `activation` is None, only check if layer can contain an activation.

    :param layer: Keras layer to check
    :type layer: Layer
    :param activation: Keras name of activation function, defaults to None
    :type activation: str, optional
    :return: If `activation` is None, check if layer contains any activation function.
        Otherwise check for specific activation function of type `activation`.
    :rtype: bool
    """

    if activation is None:
        return contains_any_activation(layer)
    a = activation.lower()

    # Check for activations in layer
    if hasattr(layer, "activation"):
        return bool(layer.activation == keras.activations.get(a))

    # Check for layers that are activations
    if a == "softmax":
        return isinstance(layer, (klayers.Softmax))
    if a == "relu":
        return isinstance(layer, (klayers.ReLU))
    if a == "elu":
        return isinstance(layer, (klayers.ELU))
    if a == "prelu":
        return isinstance(layer, (klayers.PReLU))
    if a == "leakyrelu":
        return isinstance(layer, (klayers.LeakyReLU))
    if a == "thresholdedrelu":
        return isinstance(layer, (klayers.ThresholdedReLU))
    return False


ACTIVATION_LAYERS = (
    klayers.ReLU,
    klayers.ELU,
    klayers.LeakyReLU,
    klayers.PReLU,
    klayers.Softmax,
    klayers.ThresholdedReLU,
)


def contains_any_activation(layer: Layer) -> bool:
    """Check whether layer contains any activation or is activation layer.

    :param layer: Keras layer to check
    :type layer: Layer
    :return: True if activation was found.
    :rtype: bool
    """

    return hasattr(layer, "activation") or isinstance(layer, ACTIVATION_LAYERS)


def contains_kernel(layer: Layer) -> bool:
    """
    Check whether the layer contains a kernel.
    """

    # TODO: add test and check this more throughroughly.
    # rely on Keras convention.
    return (
        hasattr(layer, "kernel")
        or hasattr(layer, "depthwise_kernel")
        or hasattr(layer, "pointwise_kernel")
    )


def only_relu_activation(layer: Layer) -> bool:
    """Checks if layer contains no or only a ReLU activation."""
    return (
        not contains_activation(layer)
        or contains_activation(layer, "linear")
        or contains_activation(layer, "relu")
    )


def is_module(layer) -> bool:
    """
    Is there a tf.Module in the network?
    """
    if isinstance(layer, klayers.Layer):
        return False
    return isinstance(layer, Module)


CONV_LAYERS = (
    klayers.Conv1D,
    klayers.Conv2D,
    klayers.Conv2DTranspose,
    klayers.Conv3D,
    klayers.Conv3DTranspose,
    klayers.SeparableConv1D,
    klayers.SeparableConv2D,
    klayers.DepthwiseConv2D,
)


def is_conv_layer(layer: Layer, *_args, **_kwargs) -> bool:
    """Checks if layer is a convolutional layer."""

    return isinstance(layer, CONV_LAYERS)


def is_embedding_layer(layer: Layer, *_args, **_kwargs) -> bool:
    """Checks if layer is an embedding layer."""
    return isinstance(layer, klayers.Embedding)


def is_batch_normalization_layer(layer: Layer, *_args, **_kwargs) -> bool:
    """Checks if layer is a batchnorm layer."""
    return isinstance(layer, klayers.BatchNormalization)


def is_add_layer(layer: Layer, *_args, **_kwargs) -> bool:
    """Checks if layer is an addition-merge layer."""
    return isinstance(layer, klayers.Add)


def is_dense_layer(layer: Layer, *_args, **_kwargs) -> bool:
    """Checks if layer is a dense layer."""
    return isinstance(layer, klayers.Dense)


CONVNET_LAYERS = (
    klayers.InputLayer,
    klayers.ELU,
    klayers.LeakyReLU,
    klayers.PReLU,
    klayers.Softmax,
    klayers.ThresholdedReLU,
    klayers.Conv1D,
    klayers.Conv2D,
    klayers.Conv2DTranspose,
    klayers.Conv3D,
    klayers.Conv3DTranspose,
    klayers.Cropping1D,
    klayers.Cropping2D,
    klayers.Cropping3D,
    klayers.SeparableConv1D,
    klayers.SeparableConv2D,
    klayers.UpSampling1D,
    klayers.UpSampling2D,
    klayers.UpSampling3D,
    klayers.ZeroPadding1D,
    klayers.ZeroPadding2D,
    klayers.ZeroPadding3D,
    klayers.Activation,
    klayers.ActivityRegularization,
    klayers.Dense,
    klayers.Dropout,
    klayers.Flatten,
    klayers.Lambda,
    klayers.Masking,
    klayers.Permute,
    klayers.RepeatVector,
    klayers.Reshape,
    klayers.SpatialDropout1D,
    klayers.SpatialDropout2D,
    klayers.SpatialDropout3D,
    klayers.Embedding,
    klayers.LocallyConnected1D,
    klayers.LocallyConnected2D,
    klayers.Add,
    klayers.Average,
    klayers.Concatenate,
    klayers.Dot,
    klayers.Maximum,
    klayers.Minimum,
    klayers.Multiply,
    klayers.Subtract,
    klayers.AlphaDropout,
    klayers.GaussianDropout,
    klayers.GaussianNoise,
    klayers.BatchNormalization,
    klayers.AveragePooling1D,
    klayers.AveragePooling2D,
    klayers.AveragePooling3D,
    klayers.GlobalAveragePooling1D,
    klayers.GlobalAveragePooling2D,
    klayers.GlobalAveragePooling3D,
    klayers.GlobalMaxPooling1D,
    klayers.GlobalMaxPooling2D,
    klayers.GlobalMaxPooling3D,
    klayers.MaxPooling1D,
    klayers.MaxPooling2D,
    klayers.MaxPooling3D,
)


def is_convnet_layer(layer: Layer) -> bool:
    """Checks if layer is from a convolutional network."""
    # Inside function to not break import if Keras changes.

    return isinstance(layer, CONVNET_LAYERS)


AVERAGEPOOLING_LAYERS = (
    klayers.AveragePooling1D,
    klayers.AveragePooling2D,
    klayers.AveragePooling3D,
    klayers.GlobalAveragePooling1D,
    klayers.GlobalAveragePooling2D,
    klayers.GlobalAveragePooling3D,
)


def is_average_pooling(layer: Layer) -> bool:
    """Checks if layer is an average-pooling layer."""

    return isinstance(layer, AVERAGEPOOLING_LAYERS)


MAXPOOLING_LAYERS = (
    klayers.MaxPooling1D,
    klayers.MaxPooling2D,
    klayers.MaxPooling3D,
    klayers.GlobalMaxPooling1D,
    klayers.GlobalMaxPooling2D,
    klayers.GlobalMaxPooling3D,
)


def is_max_pooling(layer: Layer) -> bool:
    """Checks if layer is a max-pooling layer."""
    return isinstance(layer, MAXPOOLING_LAYERS)


def get_input_layers(layer: Layer) -> set[Layer]:
    """Returns all layers that created this layer's inputs."""
    ret = set()

    for node_index in range(len(layer._inbound_nodes)):
        Xs = ibackend.to_list(layer.get_input_at(node_index))
        for X in Xs:
            ret.add(X._keras_history[0])

    return ret


def is_input_layer(layer: Layer, ignore_reshape_layers: bool = True) -> bool:
    """Checks if layer is an input layer."""
    # Triggers if ALL inputs of layer are connected
    # to a Keras input layer object.
    # Note: In the sequential api the Sequential object
    # adds the Input layer if the user does not.

    layer_inputs = get_input_layers(layer)
    # We ignore certain layers, that do not modify
    # the data content.
    # TODO: update this list!
    ignored_layers = (
        klayers.Flatten,
        klayers.Permute,
        klayers.Reshape,
    )
    while any(isinstance(x, ignored_layers) for x in layer_inputs):
        tmp = set()
        for l in layer_inputs:
            if ignore_reshape_layers and isinstance(l, ignored_layers):
                tmp.update(get_input_layers(l))
            else:
                tmp.add(l)
        layer_inputs = tmp

    return all(isinstance(x, klayers.InputLayer) for x in layer_inputs)


def is_layer_at_idx(layer: Layer, index, ignore_reshape_layers=True) -> bool:
    """Checks if layer is a layer at index index,
    by repeatedly applying is_input_layer()."""
    # TODO: implement layer index check
    raise NotImplementedError("Layer index checking hasn't been implemented yet.")



from __future__ import annotations

import inspect

import tensorflow.keras.backend as kbackend
import tensorflow.keras.layers as klayers

import innvestigate.analyzer.relevance_based.relevance_rule as rrule
import innvestigate.analyzer.relevance_based.utils as rutils
import innvestigate.backend as ibackend
import innvestigate.backend.checks as ichecks
import innvestigate.backend.graph as igraph
from innvestigate.analyzer.network_base import AnalyzerNetworkBase
from innvestigate.analyzer.reverse_base import ReverseAnalyzerBase
from innvestigate.backend.types import Layer, LayerCheck, Model, OptionalList, Tensor

__all__ = [
    "BaselineLRPZ",
    "LRP",
    "LRP_RULES",
    "LRPZ",
    "LRPEpsilon",
    "LRPWSquare",
    "LRPFlat",
    "LRPAlphaBeta",
    "LRPAlpha2Beta1",
    "LRPAlpha2Beta1IgnoreBias",
    "LRPAlpha1Beta0",
    "LRPAlpha1Beta0IgnoreBias",
    "LRPZPlus",
    "LRPZPlusFast",
    "LRPSequentialPresetA",
    "LRPSequentialPresetB",
    "LRPSequentialPresetAFlat",
    "LRPSequentialPresetBFlat",
    "LRPSequentialPresetBFlatUntilIdx",
]


###############################################################################
BASELINE_LRPZ_LAYERS = (
    klayers.InputLayer,
    klayers.Conv1D,
    klayers.Conv2D,
    klayers.Conv2DTranspose,
    klayers.Conv3D,
    klayers.Conv3DTranspose,
    klayers.Cropping1D,
    klayers.Cropping2D,
    klayers.Cropping3D,
    klayers.SeparableConv1D,
    klayers.SeparableConv2D,
    klayers.UpSampling1D,
    klayers.UpSampling2D,
    klayers.UpSampling3D,
    klayers.ZeroPadding1D,
    klayers.ZeroPadding2D,
    klayers.ZeroPadding3D,
    klayers.Activation,
    klayers.ActivityRegularization,
    klayers.Dense,
    klayers.Dropout,
    klayers.Flatten,
    klayers.Lambda,
    klayers.Masking,
    klayers.Permute,
    klayers.RepeatVector,
    klayers.Reshape,
    klayers.SpatialDropout1D,
    klayers.SpatialDropout2D,
    klayers.SpatialDropout3D,
    klayers.LocallyConnected1D,
    klayers.LocallyConnected2D,
    klayers.Add,
    klayers.Concatenate,
    klayers.Dot,
    klayers.Maximum,
    klayers.Minimum,
    klayers.Subtract,
    klayers.AlphaDropout,
    klayers.GaussianDropout,
    klayers.GaussianNoise,
    klayers.BatchNormalization,
    klayers.GlobalMaxPooling1D,
    klayers.GlobalMaxPooling2D,
    klayers.GlobalMaxPooling3D,
    klayers.MaxPooling1D,
    klayers.MaxPooling2D,
    klayers.MaxPooling3D,
)


class BaselineLRPZ(AnalyzerNetworkBase):
    """LRPZ analyzer - for testing purpose only.

    Applies the "LRP-Z" algorithm to analyze the model.
    Based on the gradient times the input formula.
    **This formula holds only for ReLU/MaxPooling networks, for which
    LRP-Z collapses into the stated formula.**

    :param model: A Keras model.
    """

    def __init__(self, model: Model, **kwargs):
        # Inside function to not break import if Keras changes.

        super().__init__(model, **kwargs)

        # Add and run model checks
        self._add_model_softmax_check()
        self._add_model_check(
            lambda layer: not ichecks.only_relu_activation(layer),
            "BaselineLRPZ only works with  ReLU activations.",
            check_type="exception",
        )
        self._add_model_check(
            lambda layer: not isinstance(layer, BASELINE_LRPZ_LAYERS),
            "BaselineLRPZ only works with a predefined set of layers.",
            check_type="exception",
        )
        self._do_model_checks()

    def _create_analysis(
        self, model: Model, stop_analysis_at_tensors: list[Tensor] = None
    ):
        if stop_analysis_at_tensors is None:
            stop_analysis_at_tensors = []

        tensors_to_analyze = [
            x
            for x in ibackend.to_list(model.inputs)
            if x not in stop_analysis_at_tensors
        ]
        gradients = ibackend.to_list(
            kbackend.gradients(model.outputs[0], tensors_to_analyze)
        )
        return [
            klayers.Multiply()([i, g]) for i, g in zip(tensors_to_analyze, gradients)
        ]


###############################################################################

# Utility list enabling name mappings via string
LRP_RULES: dict = {
    "Z": rrule.ZRule,
    "Epsilon": rrule.EpsilonRule,
    "WSquare": rrule.WSquareRule,
    "Flat": rrule.FlatRule,
    "AlphaBeta": rrule.AlphaBetaRule,
    "AlphaBetaIgnoreBias": rrule.AlphaBetaIgnoreBiasRule,
    "Alpha2Beta1": rrule.Alpha2Beta1Rule,
    "Alpha2Beta1IgnoreBias": rrule.Alpha2Beta1IgnoreBiasRule,
    "Alpha1Beta0": rrule.Alpha1Beta0Rule,
    "Alpha1Beta0IgnoreBias": rrule.Alpha1Beta0IgnoreBiasRule,
    "ZPlus": rrule.ZPlusRule,
    "ZPlusFast": rrule.ZPlusFastRule,
    "Bounded": rrule.BoundedRule,
}


class EmbeddingReverseLayer(igraph.ReverseMappingBase):
    def __init__(self, _layer, _state):
        # TODO: implement rule support.
        pass

    def apply(self, _Xs, _Ys, Rs, _reverse_state: dict):
        # the embedding layer outputs for an (indexed) input a vector.
        # thus, in the relevance backward pass, the embedding layer receives
        # relevances Rs corresponding to those vectors.
        # Due to the 1:1 relationship between input index and output mapping vector,
        # the relevance backward pass can be realized by pooling relevances
        # over the vector axis.

        # relevances are given shaped [batch_size, sequence_length, embedding_dims]
        pool_relevance = klayers.Lambda(lambda x: kbackend.sum(x, axis=-1))
        return [pool_relevance(R) for R in Rs]


class BatchNormalizationReverseLayer(igraph.ReverseMappingBase):
    """Special BN handler that applies the Z-Rule"""

    def __init__(self, layer, _state):
        config = layer.get_config()

        self._center = config["center"]
        self._scale = config["scale"]
        self._axis = config["axis"]

        self._mean = layer.moving_mean
        self._std = layer.moving_variance
        if self._center:
            self._beta = layer.beta

        # TODO: implement rule support.
        #   for BatchNormalization -> [BNEpsilon, BNAlphaBeta, BNIgnore]
        # super().__init__(layer, state)
        # how to do this:
        # super.__init__ calls select_rule and sets a self._rule class
        # check if isinstance(self_rule, EpsiloneRule), then reroute
        # to BatchNormEpsilonRule. Not pretty, but should work.

    def apply(self, Xs, Ys, Rs, _reverse_state: dict):
        input_shape = [kbackend.int_shape(x) for x in Xs]
        if len(input_shape) != 1:
            # extend below lambda layers towards multiple parameters.
            raise ValueError(
                "BatchNormalizationReverseLayer expects Xs with length 1, "
                f"but was len(Xs)={len(Xs)}"
            )
        input_shape = input_shape[0]

        # prepare broadcasting shape for layer parameters
        broadcast_shape = [1] * len(input_shape)
        broadcast_shape[self._axis] = input_shape[self._axis]
        broadcast_shape[0] = -1

        # reweight relevances as
        #        x * (y - beta)     R
        # Rin = ---------------- * ----
        #           x - mu          y
        # batch norm can be considered as 3 distinct layers of subtraction,
        # multiplication and then addition. The multiplicative scaling layer
        # has no effect on LRP and functions as a linear activation layer

        minus_mu = klayers.Lambda(
            lambda x: x - kbackend.reshape(self._mean, broadcast_shape)
        )
        minus_beta = klayers.Lambda(
            lambda x: x - kbackend.reshape(self._beta, broadcast_shape)
        )
        prepare_div = klayers.Lambda(
            lambda x: x
            + (kbackend.cast(kbackend.greater_equal(x, 0), kbackend.floatx()) * 2 - 1)
            * kbackend.epsilon()
        )

        x_minus_mu = ibackend.apply(minus_mu, Xs)
        if self._center:
            y_minus_beta = ibackend.apply(minus_beta, Ys)
        else:
            y_minus_beta = Ys

        numerator = [
            klayers.Multiply()([x, ymb, r]) for x, ymb, r in zip(Xs, y_minus_beta, Rs)
        ]
        denominator = [klayers.Multiply()([xmm, y]) for xmm, y in zip(x_minus_mu, Ys)]

        return [
            ibackend.safe_divide(n, prepare_div(d))
            for n, d in zip(numerator, denominator)
        ]


class AddReverseLayer(igraph.ReverseMappingBase):
    """Special Add layer handler that applies the Z-Rule"""

    def __init__(self, layer, _state):
        self._layer_wo_act = igraph.copy_layer_wo_activation(
            layer, name_template="reversed_kernel_%s"
        )

        # TODO: implement rule support.
        # super().__init__(layer, state)

    def apply(self, Xs, _Ys, Rs, _reverse_state: dict):
        # The outputs of the pooling operation at each location
        # is the sum of its inputs.
        # The forward message must be known in this case,
        # and are the inputs for each pooling thing.
        # The gradient is 1 for each output-to-input connection,
        # which corresponds to the "weights" of the layer.
        # It should thus be sufficient to reweight the relevances
        # and do a gradient_wrt
        # Get activations.
        Zs = ibackend.apply(self._layer_wo_act, Xs)
        # Divide incoming relevance by the activations.
        tmp = [ibackend.safe_divide(a, b) for a, b in zip(Rs, Zs)]

        # Propagate the relevance to input neurons
        # using the gradient.
        grads = ibackend.gradients(Xs, Zs, tmp)
        # Re-weight relevance with the input values.
        return [klayers.Multiply()([a, b]) for a, b in zip(Xs, grads)]


class AveragePoolingReverseLayer(igraph.ReverseMappingBase):
    """Special AveragePooling handler that applies the Z-Rule"""

    def __init__(self, layer, _state):
        self._layer_wo_act = igraph.copy_layer_wo_activation(
            layer, name_template="reversed_kernel_%s"
        )

        # TODO: implement rule support.
        # super().__init__(layer, state)

    def apply(self, Xs, _Ys, Rs, reverse_state: dict):
        # The outputs of the pooling operation at each location
        # is the sum of its inputs.
        # The forward message must be known in this case,
        # and are the inputs for each pooling thing.
        # The gradient is 1 for each output-to-input connection,
        # which corresponds to the "weights" of the layer.
        # It should thus be sufficient to reweight the relevances
        # and do a gradient_wrt
        # Get activations.
        Zs = ibackend.apply(self._layer_wo_act, Xs)
        # Divide incoming relevance by the activations.
        tmp = [ibackend.safe_divide(a, b) for a, b in zip(Rs, Zs)]

        # Propagate the relevance to input neurons
        # using the gradient.
        grads = ibackend.gradients(Xs, Zs, tmp)
        # Re-weight relevance with the input values.
        return [klayers.Multiply()([a, b]) for a, b in zip(Xs, grads)]


class LRP(ReverseAnalyzerBase):
    """
    Base class for LRP-based model analyzers


    :param model: A Keras model.

    :param rule: A rule can be a  string or a Rule object, lists thereof or
      a list of conditions [(Condition, Rule), ... ]
      gradient.

    :param input_layer_rule: either a Rule object, atuple of (low, high)
      the min/max pixel values of the inputs
    :param bn_layer_rule: either a Rule object or None.
      None means dedicated BN rule will be applied.
    """

    def __init__(
        self,
        model,
        *args,
        rule=None,
        input_layer_rule=None,
        until_layer_idx=None,
        until_layer_rule=None,
        bn_layer_rule=None,
        bn_layer_fuse_mode: str = "one_linear",
        **kwargs,
    ):
        super().__init__(model, *args, **kwargs)

        self._input_layer_rule = input_layer_rule
        self._until_layer_rule = until_layer_rule
        self._until_layer_idx = until_layer_idx
        self._bn_layer_rule = bn_layer_rule
        self._bn_layer_fuse_mode = bn_layer_fuse_mode

        if self._bn_layer_fuse_mode not in ["one_linear", "two_linear"]:
            raise ValueError(f"Unknown _bn_layer_fuse_mode {self._bn_layer_fuse_mode}")

        self._add_model_softmax_check()
        self._add_model_check(
            lambda layer: not ichecks.is_convnet_layer(layer),
            "LRP is only tested for convolutional neural networks.",
            check_type="warning",
        )

        # TODO: refactor rule type checking into separate function
        # check if rule was given explicitly.
        # rule can be a string, a list (of strings) or
        # a list of conditions [(Condition, Rule), ... ] for each layer.
        if rule is None:
            raise ValueError("Need LRP rule(s).")

        if isinstance(rule, list):
            self._rule = list(rule)
        else:
            self._rule = rule

        if isinstance(rule, str) or (
            inspect.isclass(rule) and issubclass(rule, igraph.ReverseMappingBase)
        ):  # NOTE: All LRP rules inherit from igraph.ReverseMappingBase
            # the given rule is a single string or single rule implementing cla ss
            use_conditions = True
            rules = [(lambda _: True, rule)]

        elif not isinstance(rule[0], tuple):
            # rule list of rule strings or classes
            use_conditions = False
            rules = list(rule)
        else:
            # rule is list of conditioned rules
            use_conditions = True
            rules = rule

        # apply rule to first self._until_layer_idx layers
        if self._until_layer_rule is not None and self._until_layer_idx is not None:
            for i in range(self._until_layer_idx + 1):
                is_at_idx: LayerCheck = lambda layer: ichecks.is_layer_at_idx(layer, i)
                rules.insert(0, (is_at_idx, self._until_layer_rule))

        # create a BoundedRule for input layer handling from given tuple
        if self._input_layer_rule is not None:
            input_layer_rule = self._input_layer_rule
            if isinstance(input_layer_rule, tuple):
                low, high = input_layer_rule

                class BoundedProxyRule(rrule.BoundedRule):
                    def __init__(self, *args, **kwargs):
                        super().__init__(*args, low=low, high=high, **kwargs)

                input_layer_rule = BoundedProxyRule

            if use_conditions is True:
                rules.insert(0, (ichecks.is_input_layer, input_layer_rule))
            else:
                rules.insert(0, input_layer_rule)

        self._rules_use_conditions = use_conditions
        self._rules = rules

    def create_rule_mapping(self, layer: Layer, reverse_state: dict):
        if self._rules_use_conditions is True:
            for condition, rule in self._rules:
                if condition(layer):
                    rule_class = rule
                    break
        else:
            rule_class = self._rules.pop()

        if rule_class is None:
            raise Exception(f"No rule applies to layer {layer}")

        if isinstance(rule_class, str):
            rule_class = LRP_RULES[rule_class]
        rule = rule_class(layer, reverse_state)

        return rule.apply

    def _create_analysis(self, *args, **kwargs):
        ###################################################################
        # Functionality responible for backwards rule selection below  ####
        ###################################################################

        # default backward hook
        self._add_conditional_reverse_mapping(
            ichecks.contains_kernel,
            self.create_rule_mapping,
            name="lrp_layer_with_kernel_mapping",
        )

        # specialized backward hooks.
        # TODO: add ReverseLayer class handling layers without kernel: Add and AvgPool
        bn_layer_rule = self._bn_layer_rule

        if bn_layer_rule is None:
            # TODO (alber): get rid of this option!
            # alternatively a default rule should be applied.
            bn_mapping = BatchNormalizationReverseLayer
        else:
            if isinstance(bn_layer_rule, str):
                bn_layer_rule = LRP_RULES[bn_layer_rule]

            bn_mapping = igraph.apply_mapping_to_fused_bn_layer(
                bn_layer_rule,
                fuse_mode=self._bn_layer_fuse_mode,
            )
        self._add_conditional_reverse_mapping(
            ichecks.is_batch_normalization_layer,
            bn_mapping,
            name="lrp_batch_norm_mapping",
        )
        self._add_conditional_reverse_mapping(
            ichecks.is_average_pooling,
            AveragePoolingReverseLayer,
            name="lrp_average_pooling_mapping",
        )
        self._add_conditional_reverse_mapping(
            ichecks.is_add_layer,
            AddReverseLayer,
            name="lrp_add_layer_mapping",
        )
        self._add_conditional_reverse_mapping(
            ichecks.is_embedding_layer,
            EmbeddingReverseLayer,
            name="lrp_embedding_mapping",
        )

        # FINALIZED constructor.
        return super()._create_analysis(*args, **kwargs)

    def _default_reverse_mapping(
        self,
        Xs: OptionalList[Tensor],
        Ys: OptionalList[Tensor],
        reversed_Ys: OptionalList[Tensor],
        reverse_state: dict,
    ):
        # default_return_layers = [klayers.Activation]# TODO extend
        if (
            len(Xs) == len(Ys)
            and isinstance(reverse_state["layer"], (klayers.Activation,))
            and all(
                kbackend.int_shape(x) == kbackend.int_shape(y) for x, y in zip(Xs, Ys)
            )
        ):
            # Expect Xs and Ys to have the same shapes.
            # There is not mixing of relevances as there is kernel,
            # therefore we pass them as they are.
            return reversed_Ys

        # This branch covers:
        # MaxPooling, Max, Flatten, Reshape, Concatenate, Cropping
        return self._gradient_reverse_mapping(Xs, Ys, reversed_Ys, reverse_state)

    ######################################
    # End of Rule Selection Business. ####
    ######################################

    def _get_state(self):
        state = super()._get_state()
        state.update({"rule": self._rule})
        state.update({"input_layer_rule": self._input_layer_rule})
        state.update({"bn_layer_rule": self._bn_layer_rule})
        state.update({"bn_layer_fuse_mode": self._bn_layer_fuse_mode})
        return state

    @classmethod
    def _state_to_kwargs(cls, state):
        rule = state.pop("rule")
        input_layer_rule = state.pop("input_layer_rule")
        bn_layer_rule = state.pop("bn_layer_rule")
        bn_layer_fuse_mode = state.pop("bn_layer_fuse_mode")
        # call super after popping class-specific states:
        kwargs = super()._state_to_kwargs(state)

        kwargs.update(
            {
                "rule": rule,
                "input_layer_rule": input_layer_rule,
                "bn_layer_rule": bn_layer_rule,
                "bn_layer_fuse_mode": bn_layer_fuse_mode,
            }
        )
        return kwargs


###############################################################################
# ANALYZER CLASSES AND PRESETS ################################################
###############################################################################


class _LRPFixedParams(LRP):
    @classmethod
    def _state_to_kwargs(cls, state):
        # call super after popping class-specific states:
        kwargs = super()._state_to_kwargs(state)

        del kwargs["rule"]
        del kwargs["bn_layer_rule"]
        return kwargs


class LRPZ(_LRPFixedParams):
    """LRP-analyzer that uses the LRP-Z rule"""

    def __init__(self, model, *args, **kwargs):
        super().__init__(model, *args, rule="Z", bn_layer_rule="Z", **kwargs)
        self._do_model_checks()


class LRPEpsilon(_LRPFixedParams):
    """LRP-analyzer that uses the LRP-Epsilon rule"""

    def __init__(self, model, epsilon=1e-7, bias=True, *args, **kwargs):
        epsilon = rutils.assert_lrp_epsilon_param(epsilon, self)
        self._epsilon = epsilon

        class EpsilonProxyRule(rrule.EpsilonRule):
            """
            Dummy class inheriting from EpsilonRule
            for passing along the chosen parameters from
            the LRP analyzer class to the decopmosition rules.
            """

            def __init__(self, *args, **kwargs):
                super().__init__(*args, epsilon=epsilon, bias=bias, **kwargs)

        super().__init__(
            model,
            *args,
            rule=EpsilonProxyRule,
            bn_layer_rule=EpsilonProxyRule,
            **kwargs,
        )

        self._do_model_checks()


class LRPWSquare(_LRPFixedParams):
    """LRP-analyzer that uses the DeepTaylor W**2 rule"""

    def __init__(self, model, *args, **kwargs):
        super().__init__(
            model, *args, rule="WSquare", bn_layer_rule="WSquare", **kwargs
        )
        self._do_model_checks()


class LRPFlat(_LRPFixedParams):
    """LRP-analyzer that uses the LRP-Flat rule"""

    def __init__(self, model, *args, **kwargs):
        super().__init__(model, *args, rule="Flat", bn_layer_rule="Flat", **kwargs)
        self._do_model_checks()


class LRPAlphaBeta(LRP):
    """Base class for LRP AlphaBeta"""

    def __init__(self, model, *args, alpha=None, beta=None, bias=True, **kwargs):
        alpha, beta = rutils.assert_infer_lrp_alpha_beta_param(alpha, beta, self)
        self._alpha = alpha
        self._beta = beta
        self._bias = bias

        class AlphaBetaProxyRule(rrule.AlphaBetaRule):
            """
            Dummy class inheriting from AlphaBetaRule
            for the purpose of passing along the chosen parameters from
            the LRP analyzer class to the decopmosition rules.
            """

            def __init__(self, *args, **kwargs):
                super().__init__(*args, alpha=alpha, beta=beta, bias=bias, **kwargs)

        super().__init__(
            model,
            *args,
            rule=AlphaBetaProxyRule,
            bn_layer_rule=AlphaBetaProxyRule,
            **kwargs,
        )
        self._do_model_checks()

    def _get_state(self):
        state = super()._get_state()
        del state["rule"]
        state.update({"alpha": self._alpha})
        state.update({"beta": self._beta})
        state.update({"bias": self._bias})
        return state

    @classmethod
    def _state_to_kwargs(cls, state):
        alpha = state.pop("alpha")
        beta = state.pop("beta")
        bias = state.pop("bias")
        state["rule"] = None
        # call super after popping class-specific states:
        kwargs = super()._state_to_kwargs(state)

        del kwargs["rule"]
        del kwargs["bn_layer_rule"]
        kwargs.update({"alpha": alpha, "beta": beta, "bias": bias})
        return kwargs


class _LRPAlphaBetaFixedParams(LRPAlphaBeta):
    @classmethod
    def _state_to_kwargs(cls, state):
        # call super after popping class-specific states:
        kwargs = super()._state_to_kwargs(state)

        del kwargs["alpha"]
        del kwargs["beta"]
        del kwargs["bias"]
        return kwargs


class LRPAlpha2Beta1(_LRPAlphaBetaFixedParams):
    """LRP-analyzer that uses the LRP-alpha-beta rule with a=2,b=1"""

    def __init__(self, model, *args, **kwargs):
        super().__init__(model, *args, alpha=2, beta=1, bias=True, **kwargs)
        self._do_model_checks()


class LRPAlpha2Beta1IgnoreBias(_LRPAlphaBetaFixedParams):
    """LRP-analyzer that uses the LRP-alpha-beta-ignbias rule with a=2,b=1"""

    def __init__(self, model, *args, **kwargs):
        super().__init__(model, *args, alpha=2, beta=1, bias=False, **kwargs)
        self._do_model_checks()


class LRPAlpha1Beta0(_LRPAlphaBetaFixedParams):
    """LRP-analyzer that uses the LRP-alpha-beta rule with a=1,b=0"""

    def __init__(self, model, *args, **kwargs):
        super().__init__(model, *args, alpha=1, beta=0, bias=True, **kwargs)
        self._do_model_checks()


class LRPAlpha1Beta0IgnoreBias(_LRPAlphaBetaFixedParams):
    """LRP-analyzer that uses the LRP-alpha-beta-ignbias rule with a=1,b=0"""

    def __init__(self, model, *args, **kwargs):
        super().__init__(model, *args, alpha=1, beta=0, bias=False, **kwargs)
        self._do_model_checks()


class LRPZPlus(LRPAlpha1Beta0IgnoreBias):
    """LRP-analyzer that uses the LRP-alpha-beta rule with a=1,b=0"""

    # TODO: assert that layer inputs are always >= 0
    def __init__(self, model, *args, **kwargs):
        super().__init__(model, *args, **kwargs)
        self._do_model_checks()


class LRPZPlusFast(_LRPFixedParams):
    """
    The ZPlus rule is a special case of the AlphaBetaRule
    for alpha=1, beta=0 and assumes inputs x >= 0.
    """

    # TODO: assert that layer inputs are always >= 0
    def __init__(self, model, *args, **kwargs):
        super().__init__(
            model, *args, rule="ZPlusFast", bn_layer_rule="ZPlusFast", **kwargs
        )
        self._do_model_checks()


class LRPSequentialPresetA(_LRPFixedParams):  # for the lack of a better name
    """Special LRP-configuration for ConvNets"""

    def __init__(
        self,
        model,
        *args,
        epsilon=1e-1,
        bn_layer_rule=rrule.AlphaBetaX2m100Rule,
        **kwargs,
    ):
        class EpsilonProxyRule(rrule.EpsilonRule):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, epsilon=epsilon, bias=True, **kwargs)

        conditional_rules = [
            (ichecks.is_dense_layer, EpsilonProxyRule),
            (ichecks.is_conv_layer, rrule.Alpha1Beta0Rule),
        ]

        super().__init__(
            model, *args, rule=conditional_rules, bn_layer_rule=bn_layer_rule, **kwargs
        )

        self._add_model_check(
            lambda layer: not ichecks.only_relu_activation(layer),
            # TODO: fix. specify. extend.
            (
                "LRPSequentialPresetA is not advised "
                "for networks with non-ReLU activations."
            ),
            check_type="warning",
        )

        self._do_model_checks()


class LRPSequentialPresetB(_LRPFixedParams):
    """Special LRP-configuration for ConvNets"""

    def __init__(
        self,
        model: Model,
        *args,
        epsilon: float = 1e-1,
        bn_layer_rule=rrule.AlphaBetaX2m100Rule,
        **kwargs,
    ):
        class EpsilonProxyRule(rrule.EpsilonRule):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, epsilon=epsilon, bias=True, **kwargs)

        conditional_rules = [
            (ichecks.is_dense_layer, EpsilonProxyRule),
            (ichecks.is_conv_layer, rrule.Alpha2Beta1Rule),
        ]

        super().__init__(
            model, *args, rule=conditional_rules, bn_layer_rule=bn_layer_rule, **kwargs
        )

        # Add and run model checks
        self._add_model_check(
            lambda layer: not ichecks.only_relu_activation(layer),
            # TODO: fix. specify. extend.
            (
                "LRPSequentialPresetB is not advised "
                "for networks with non-ReLU activations."
            ),
            check_type="warning",
        )
        self._do_model_checks()


# TODO: allow to pass input layer identification by index or id.
class LRPSequentialPresetAFlat(LRPSequentialPresetA):
    """Special LRP-configuration for ConvNets"""

    def __init__(self, model, *args, **kwargs):
        # provide functionality for `analyzer.load()` by avoiding multiple kwargs:
        if "input_layer_rule" in kwargs:
            if kwargs["input_layer_rule"] != "Flat":
                raise RuntimeError(
                    "Unexpected input_layer_rule when loading LRPSequentialPresetAFlat."
                )
            kwargs.pop("input_layer_rule")
        super().__init__(model, *args, input_layer_rule="Flat", **kwargs)


# TODO: allow to pass input layer identification by index or id.
class LRPSequentialPresetBFlat(LRPSequentialPresetB):
    """Special LRP-configuration for ConvNets"""

    def __init__(self, model, *args, **kwargs):
        # provide functionality for `analyzer.load()` by avoiding multiple kwargs:
        if "input_layer_rule" in kwargs:
            if kwargs["input_layer_rule"] != "Flat":
                raise RuntimeError(
                    "Unexpected input_layer_rule when loading LRPSequentialPresetBFlat."
                )
            kwargs.pop("input_layer_rule")
        super().__init__(model, *args, input_layer_rule="Flat", **kwargs)


class LRPSequentialPresetBFlatUntilIdx(LRPSequentialPresetBFlat):
    """
    Special LRP-configuration for ConvNets.
    Allows to perform LRP_flat from (including) layer until_layer_idx down until
    the input layer. Weightless layers are ignored when counting the index for now.
    """

    def __init__(self, model, *args, until_layer_idx=None, **kwargs):
        super().__init__(
            model,
            *args,
            until_layer_idx=until_layer_idx,
            until_layer_rule=rrule.FlatRule,
            **kwargs,
        )



from __future__ import annotations

from typing import Any

import numpy as np
import tensorflow.keras.backend as kbackend
import tensorflow.keras.layers as klayers
import tensorflow.keras.models as kmodels

import innvestigate.backend as ibackend
import innvestigate.backend.checks as ichecks
import innvestigate.layers as ilayers
from innvestigate.analyzer.base import AnalyzerBase
from innvestigate.backend.types import Layer, LayerCheck, Model, OptionalList, Tensor

__all__ = ["AnalyzerNetworkBase"]


class AnalyzerNetworkBase(AnalyzerBase):
    """Convenience interface for analyzers.

    This class provides helpful functionality to create analyzer's.
    Basically it:

    * takes the input model and adds a layer that selects
      the desired output neuron to analyze.
    * passes the new model to :func:`_create_analysis` which should
      return the analysis as Keras tensors.
    * compiles the function and serves the output to :func:`analyze` calls.
    * allows :func:`_create_analysis` to return tensors
      that are intercept for debugging purposes.

    :param allow_lambda_layers: Allow the model to contain lambda layers.
    """

    def __init__(
        self,
        model: Model,
        allow_lambda_layers: bool = False,
        **kwargs,
    ) -> None:
        """
        From AnalyzerBase super init:
        * Initializes empty list of _model_checks
        * set _neuron_selection_mode

        Here:
        * add check for lambda layers through 'allow_lambda_layers'
        * define attributes for '_prepare_model', which is later called
            through 'create_analyzer_model'
        """
        # Call super init to initialize self._model_checks
        super().__init__(model, **kwargs)

        # Add model check for lambda layers
        self._allow_lambda_layers: bool = allow_lambda_layers
        self._add_lambda_layers_check()

        # Attributes of prepared model created by '_prepare_model'
        self._analyzer_model_done: bool = False
        self._analyzer_model: Model = None
        self._special_helper_layers: list[Layer] = []  # added for _reverse_mapping
        self._analysis_inputs: list[Tensor] | None = None
        self._n_data_input: int = 0
        self._n_constant_input: int = 0
        self._n_data_output: int = 0
        self._n_debug_output: int = 0

    def _add_model_softmax_check(self) -> None:
        """
        Adds check that prevents models from containing a softmax.
        """
        contains_softmax: LayerCheck = lambda layer: ichecks.contains_activation(
            layer, activation="softmax"
        )
        self._add_model_check(
            check=contains_softmax,
            message="This analysis method does not support softmax layers.",
            check_type="exception",
        )

    def _add_lambda_layers_check(self) -> None:
        check_lambda_layers: LayerCheck = lambda layer: (
            not self._allow_lambda_layers and isinstance(layer, klayers.Lambda)
        )
        self._add_model_check(
            check=check_lambda_layers,
            message=(
                "Lambda layers are not allowed. "
                "To force use set 'allow_lambda_layers' parameter."
            ),
            check_type="exception",
        )

    def _prepare_model(self, model: Model) -> tuple[Model, list[Tensor], list[Tensor]]:
        """
        Prepares the model to analyze before it gets actually analyzed.

        This class adds the code to select a specific output neuron.
        """
        neuron_selection_mode: str
        model_inputs: list[Tensor]
        model_output: list[Tensor]
        analysis_inputs: list[Tensor]
        stop_analysis_at_tensors: list[Tensor]

        neuron_selection_mode = self._neuron_selection_mode
        model_inputs = model.inputs
        model_output = model.outputs

        if len(model_output) > 1:
            raise ValueError("Only models with one output tensor are allowed.")

        analysis_inputs = []
        stop_analysis_at_tensors = []

        # Flatten to form (batch_size, other_dimensions):
        if kbackend.ndim(model_output[0]) > 2:
            model_output = klayers.Flatten()(model_output)

        if neuron_selection_mode == "max_activation":
            inn_max = ilayers.MaxNeuronSelection(name="MaxNeuronSelection")
            model_output = inn_max(model_output)
            self._special_helper_layers.append(inn_max)

        elif neuron_selection_mode == "index":
            # Creates a placeholder tensor when `dtype` is passed.
            neuron_indexing: Layer = klayers.Input(
                shape=(2,),  # infer amount of output neurons
                dtype=np.int32,
                name="iNNvestigate_neuron_indexing",
            )
            # TODO: what does _keras_history[0] do?
            self._special_helper_layers.append(neuron_indexing._keras_history[0])
            analysis_inputs.append(neuron_indexing)
            stop_analysis_at_tensors.append(neuron_indexing)

            select = ilayers.NeuronSelection(name="NeuronSelection")
            model_output = select(model_output + [neuron_indexing])
            self._special_helper_layers.append(select)
        elif neuron_selection_mode == "all":
            pass
        else:
            raise NotImplementedError()

        inputs = model_inputs + analysis_inputs
        outputs = model_output
        model = kmodels.Model(inputs=inputs, outputs=outputs)

        return model, analysis_inputs, stop_analysis_at_tensors

    def create_analyzer_model(self) -> None:
        """
        Creates the analyze functionality. If not called beforehand
        it will be called by :func:`analyze`.
        """
        model_inputs: list[Tensor] = self._model.inputs
        model, analysis_inputs, stop_analysis_at_tensors = self._prepare_model(
            self._model
        )
        self._analysis_inputs = analysis_inputs
        self._prepared_model = model

        tmp = self._create_analysis(
            model, stop_analysis_at_tensors=stop_analysis_at_tensors
        )
        if isinstance(tmp, tuple):
            if len(tmp) == 3:
                analysis_outputs, debug_outputs, constant_inputs = tmp  # type: ignore
            elif len(tmp) == 2:
                analysis_outputs, debug_outputs = tmp  # type: ignore
                constant_inputs = []
            elif len(tmp) == 1:
                analysis_outputs = tmp[0]
                constant_inputs = []
                debug_outputs = []
            else:
                raise Exception("Unexpected output from _create_analysis.")
        else:
            analysis_outputs = tmp
            constant_inputs = []
            debug_outputs = []

        analysis_outputs = ibackend.to_list(analysis_outputs)
        debug_outputs = ibackend.to_list(debug_outputs)
        constant_inputs = ibackend.to_list(constant_inputs)

        self._n_data_input = len(model_inputs)
        self._n_constant_input = len(constant_inputs)
        self._n_data_output = len(analysis_outputs)
        self._n_debug_output = len(debug_outputs)

        inputs = model_inputs + analysis_inputs + constant_inputs
        outputs = analysis_outputs + debug_outputs

        self._analyzer_model = kmodels.Model(
            inputs=inputs,
            outputs=outputs,
            name=f"{self.__class__.__name__}_analyzer_model",
        )
        self._analyzer_model_done = True

    def _create_analysis(
        self, model: Model, stop_analysis_at_tensors: list[Tensor] = None
    ) -> (
        tuple[list[Tensor]]
        | tuple[list[Tensor], list[Tensor]]
        | tuple[list[Tensor], list[Tensor], list[Tensor]]
    ):
        """
        Interface that needs to be implemented by a derived class.

        This function is expected to create a Keras graph that creates
        a custom analysis for the model inputs given the model outputs.

        :param model: Target of analysis.
        :param stop_analysis_at_tensors: A list of tensors where to stop the
          analysis. Similar to stop_gradient arguments when computing the
          gradient of a graph.
        :return: Either one-, two- or three-tuple of lists of tensors.
          * The first list of tensors represents the analysis for each
            model input tensor. Tensors present in stop_analysis_at_tensors
            should be omitted.
          * The second list, if present, is a list of debug tensors that will
            be passed to :func:`_handle_debug_output` after the analysis
            is executed.
          * The third list, if present, is a list of constant input tensors
            added to the analysis model.
        """
        raise NotImplementedError()

    def _handle_debug_output(self, debug_values):
        raise NotImplementedError()

    def analyze(
        self,
        X: OptionalList[np.ndarray],
        neuron_selection: OptionalList[int] | None = None,
    ) -> OptionalList[np.ndarray]:
        """
        Same interface as :class:`Analyzer` besides

        :param neuron_selection: If neuron_selection_mode is 'index' this
        should be an integer with the index for the chosen neuron.
        When analyzing batches, this should be a List of integer indices.
        """
        # TODO: what does should mean in docstring?

        if self._analyzer_model_done is False:
            self.create_analyzer_model()

        if neuron_selection is not None and self._neuron_selection_mode != "index":
            raise ValueError(
                f"neuron_selection_mode {self._neuron_selection_mode} doesn't support ",
                "'neuron_selection' parameter.",
            )

        if neuron_selection is None and self._neuron_selection_mode == "index":
            raise ValueError(
                "neuron_selection_mode 'index' expects 'neuron_selection' parameter."
            )

        ret: OptionalList[np.ndarray]
        if self._neuron_selection_mode == "index":
            if neuron_selection is not None:
                batch_size = np.shape(X)[0]
                indices = self._get_neuron_selection_array(neuron_selection, batch_size)
                ret = self._analyzer_model.predict_on_batch([X, indices])
            else:
                raise RuntimeError(
                    'neuron_selection_mode "index" requires neuron_selection.'
                )
        else:
            ret = self._analyzer_model.predict_on_batch(X)

        if self._n_debug_output > 0:
            self._handle_debug_output(ret[-self._n_debug_output :])
            ret = ret[: -self._n_debug_output]

        return ibackend.unpack_singleton(ret)

    def _get_neuron_selection_array(
        self, neuron_selection: OptionalList[int], batch_size: int
    ) -> np.ndarray:
        """Get neuron selection array for neuron_selection_mode "index"."""
        nsa = np.asarray(neuron_selection).flatten()

        # If `nsa` is singleton, repeat it so that it matches the batch size
        if nsa.size == 1:
            nsa = np.repeat(nsa, batch_size)

        # Multiples of batch size are allowed for use with AugmentReduceBase:
        if nsa.size % batch_size != 0:
            raise ValueError(
                f"""`neuron_selection` should be integer or array matching
                batch size {batch_size}. Got: {neuron_selection}"""
            )

        # We prepend a counter for the position in the batch,
        # which will be used by the layer `NeuronSelection`.
        # Using `nsa.size` for compatibility with AugmentReduceBase.
        batch_position_index = np.arange(nsa.size)
        return np.hstack((batch_position_index.reshape((-1, 1)), nsa.reshape((-1, 1))))

    def _get_state(self) -> dict[str, Any]:
        state = super()._get_state()
        state.update({"allow_lambda_layers": self._allow_lambda_layers})
        return state

    @classmethod
    def _state_to_kwargs(cls, state: dict[str, Any]) -> dict[str, Any]:
        allow_lambda_layers = state.pop("allow_lambda_layers")
        # call super after popping class-specific states:
        kwargs = super()._state_to_kwargs(state)

        kwargs.update({"allow_lambda_layers": allow_lambda_layers})
        return kwargs



from __future__ import annotations

import math
from typing import Callable

import tensorflow.keras.utils as kutils

from innvestigate.backend import to_list
from innvestigate.backend.types import OptionalList, Tensor

__all__ = [
    "BatchSequence",
    "TargetAugmentedSequence",
]


class BatchSequence(kutils.Sequence):
    """Batch sequence generator.

    Take a (list of) input tensors and a batch size
    and creates a generators that creates a sequence of batches.

    :param Xs: One or a list of tensors. First axis needs to have same length.
    :param batch_size: Batch size. Default 32.
    """

    def __init__(self, Xs: OptionalList[Tensor], batch_size: int = 32) -> None:
        self.Xs: list[Tensor] = to_list(Xs)
        self.single_tensor: bool = len(Xs) == 1
        self.batch_size: int = batch_size

        if not self.single_tensor:
            for X in self.Xs[1:]:
                assert X.shape[0] == self.Xs[0].shape[0]
        super().__init__()

    def __len__(self) -> int:
        return int(math.ceil(float(len(self.Xs[0])) / self.batch_size))

    def __getitem__(self, idx: int) -> Tensor | tuple[Tensor]:
        ret: list[Tensor] = [
            X[idx * self.batch_size : (idx + 1) * self.batch_size] for X in self.Xs
        ]

        if self.single_tensor:
            return ret[0]
        return tuple(ret)


class TargetAugmentedSequence(kutils.Sequence):
    """Augments a sequence with a target on the fly.

    Takes a sequence/generator and a function that
    creates on the fly for each batch a target.
    The generator takes a batch from that sequence,
    computes the target and returns both.

    :param sequence: A sequence or generator.
    :param augment_f: Takes a batch and returns a target.
    """

    def __init__(
        self, sequence: list[Tensor], augment_f: Callable[[list[Tensor]], list[Tensor]]
    ) -> None:
        self.sequence = sequence
        self.augment_f = augment_f

        super().__init__()

    def __len__(self) -> int:
        return len(self.sequence)

    def __getitem__(self, idx: int) -> tuple[list[Tensor], list[Tensor]]:
        inputs = self.sequence[idx]
        if isinstance(inputs, tuple):  # TODO: check if this can be removed
            assert len(inputs) == 1
            inputs = inputs[0]

        targets = self.augment_f(to_list(inputs))
        return inputs, targets



from __future__ import annotations

import warnings
from abc import ABCMeta, abstractmethod
from typing import Any

import numpy as np
import tensorflow.keras.models as kmodels

import innvestigate.analyzer
import innvestigate.backend.graph as igraph
import innvestigate.utils.sequence as isequence
from innvestigate.backend.types import LayerCheck, Model, ModelCheckDict, OptionalList

__all__ = [
    "NotAnalyzeableModelException",
    "AnalyzerBase",
    "TrainerMixin",
    "OneEpochTrainerMixin",
]


class NotAnalyzeableModelException(Exception):
    """Indicates that the model cannot be analyzed by an analyzer."""


class AnalyzerBase(metaclass=ABCMeta):
    """The basic interface of an iNNvestigate analyzer.

    This class defines the basic interface for analyzers:

    >>> model = create_keras_model()
    >>> a = Analyzer(model)
    >>> analysis = a.analyze(X_test)
    >>>
    >>> state = a.save()
    >>> a_new = A.load(*state)
    >>> analysis = a_new.analyze(X_test)

    :param model: A Keras model.
    :param disable_model_checks: Do not execute model checks that enforce
      compatibility of analyzer and model.
    :param neuron_selection_mode: How to select the neuron to analyze.
      Possible values are 'max_activation', 'index' for the neuron
      (expects indices at :func:`analyze` calls), 'all' take all neurons.

    .. note:: To develop a new analyzer derive from
      :class:`AnalyzerNetworkBase`.
    """

    def __init__(
        self,
        model: Model,
        neuron_selection_mode: str = "max_activation",
        disable_model_checks: bool = False,
        _model_check_done: bool = False,
        _model_checks: list[ModelCheckDict] = None,
    ) -> None:
        self._model = model
        self._disable_model_checks = disable_model_checks
        self._model_check_done = _model_check_done

        # There are three possible neuron selection modes
        # that return an explanation w.r.t.:
        # * "max_activation": maximum activated neuron
        # * "index": neuron at index given on call to `analyze`
        # * "all": all output neurons
        if neuron_selection_mode not in ["max_activation", "index", "all"]:
            raise ValueError("neuron_selection_mode parameter is not valid.")
        self._neuron_selection_mode: str = neuron_selection_mode

        # If no model checks are given, initialize an empty list of checks
        # that child analyzers can append to.
        if _model_checks is None:
            _model_checks = []
        self._model_checks: list[ModelCheckDict] = _model_checks

    def _add_model_check(
        self, check: LayerCheck, message: str, check_type: str = "exception"
    ) -> None:
        """Add model check to list of checks `self._model_checks`.

        :param check: Callable that performs a boolean check on a Keras layers.
        :type check: LayerCheck
        :param message: Error message if check fails.
        :type message: str
        :param check_type: Either "exception" or "warning". Defaults to "exception"
        :type check_type: str, optional
        :raises Exception: [description]
        """

        if self._model_check_done:
            raise Exception(
                "Cannot add model check anymore. Check was already performed."
            )

        check_instance: ModelCheckDict = {
            "check": check,
            "message": message,
            "check_type": check_type,
        }
        self._model_checks.append(check_instance)

    def _do_model_checks(self) -> None:
        if not self._disable_model_checks and len(self._model_checks) > 0:
            check = [x["check"] for x in self._model_checks]
            types = [x["check_type"] for x in self._model_checks]
            messages = [x["message"] for x in self._model_checks]

            checked = igraph.model_contains(self._model, check)

            tmp = zip(checked, messages, types)

            for checked_layers, message, check_type in tmp:
                if len(checked_layers) > 0:
                    tmp_message = (
                        message + f"\nCheck triggered by layers: {checked_layers}"
                    )

                    if check_type == "exception":
                        raise NotAnalyzeableModelException(tmp_message)
                    if check_type == "warning":
                        # TODO(albermax) only the first warning will be shown
                        warnings.warn(tmp_message)
                    raise NotImplementedError()

        self._model_check_done = True

    @abstractmethod
    def analyze(
        self, X: OptionalList[np.ndarray], *args: Any, **kwargs: Any
    ) -> OptionalList[np.ndarray]:
        """
        Analyze the behavior of model on input `X`.

        :param X: Input as expected by model.
        """

    def _get_state(self) -> dict:
        state = {
            "model_json": self._model.to_json(),
            "model_weights": self._model.get_weights(),
            "disable_model_checks": self._disable_model_checks,
            "neuron_selection_mode": self._neuron_selection_mode,
        }
        return state

    def save(self) -> tuple[str, dict]:
        """
        Save state of analyzer, can be passed to :func:`Analyzer.load`
        to resemble the analyzer.

        :return: The class name and the state.
        """
        state = self._get_state()
        class_name = self.__class__.__name__
        return class_name, state

    def save_npz(self, fname: str) -> None:
        """
        Save state of analyzer, can be passed to :func:`Analyzer.load_npz`
        to resemble the analyzer.

        :param fname: The file's name.
        """
        class_name, state = self.save()
        np.savez(fname, **{"class_name": class_name, "state": state})

    @classmethod
    def _state_to_kwargs(cls, state: dict) -> dict:
        disable_model_checks = state.pop("disable_model_checks")
        model_json = state.pop("model_json")
        model_weights = state.pop("model_weights")
        neuron_selection_mode = state.pop("neuron_selection_mode")

        # since `super()._state_to_kwargs(state)` should be called last
        # in every child class, the dict `state` should be empty at this point.
        if len(state) != 0:
            raise RuntimeError(f"Serialization failed. Got left-over state {state}.")

        model = kmodels.model_from_json(model_json)
        model.set_weights(model_weights)
        return {
            "model": model,
            "disable_model_checks": disable_model_checks,
            "neuron_selection_mode": neuron_selection_mode,
        }

    @staticmethod
    def load(class_name: str, state: dict[str, Any]) -> AnalyzerBase:
        """
        Resembles an analyzer from the state created by
        :func:`analyzer.save()`.

        :param class_name: The analyzer's class name.
        :param state: The analyzer's state.
        """
        # TODO: do in a smarter way without cyclic imports
        cls = getattr(innvestigate.analyzer, class_name)

        kwargs = cls._state_to_kwargs(state)
        return cls(**kwargs)  # type: ignore

    @staticmethod
    def load_npz(fname):
        """
        Resembles an analyzer from the file created by
        :func:`analyzer.save_npz()`.

        :param fname: The file's name.
        """
        npz_file = np.load(fname)

        class_name = npz_file["class_name"].item()
        state = npz_file["state"].item()
        return AnalyzerBase.load(class_name, state)



from __future__ import annotations

__all__ = ["assert_lrp_epsilon_param", "assert_infer_lrp_alpha_beta_param"]


def assert_lrp_epsilon_param(epsilon, caller):
    """
    Function for asserting epsilon parameter choice
    passed to constructors inheriting from EpsilonRule
    and LRPEpsilon.
    The following conditions can not be met:

    epsilon > 1

    :param epsilon: the epsilon parameter.
    :param caller: the class instance calling this assertion function
    """

    if epsilon <= 0:
        err_head = f"Constructor call to {caller.__class__.__name__} : "
        err_msg = err_head + f"Parameter epsilon must be > 0 but was {epsilon}"
        raise ValueError(err_msg)
    return epsilon


def assert_infer_lrp_alpha_beta_param(alpha, beta, caller):
    """
    Function for asserting parameter choices for alpha and beta
    passed to constructors inheriting from AlphaBetaRule
    and LRPAlphaBeta.

    since alpha - beta are subjected to sum to 1,
    it is sufficient for only one of the parameters to be passed
    to a corresponding class constructor.
    this method will cause an assertion error if both are None
    or the following conditions can not be met

    alpha >= 1
    beta >= 0
    alpha - beta = 1

    :param alpha: the alpha parameter.
    :param beta: the beta parameter
    :param caller: the class instance calling this assertion function
    """

    # TODO: Rework error messages

    err_head = f"Constructor call to {caller.__class__.__name__} : "
    if alpha is None and beta is None:
        err_msg = err_head + "Neither alpha or beta were given"
        raise ValueError(err_msg)

    # assert passed parameter choices
    if alpha is not None and alpha < 1:
        err_msg = (
            err_head
            + f"Passed parameter alpha invalid. Expecting alpha >= 1 but was {alpha}"
        )
        raise ValueError(err_msg)

    if beta is not None and beta < 0:
        err_msg = (
            err_head
            + f"Passed parameter beta invalid. Expecting beta >= 0 but was {beta}"
        )
        raise ValueError(err_msg)

    # assert inferred parameter choices
    if alpha is None:
        alpha = beta + 1
        if alpha < 1:
            err_msg = (
                err_head
                + f"Inferring alpha from given beta {beta} s.t. alpha - beta = 1, "
                + "with condition alpha >= 1 not possible."
            )
            raise ValueError(err_msg)

    if beta is None:
        beta = alpha - 1
        if beta < 0:
            err_msg = (
                err_head
                + f"Inferring beta from given alpha {alpha} s.t. alpha - beta = 1, "
                + "with condition beta >= 0 not possible."
            )
            raise ValueError(err_msg)

    # final check: alpha - beta = 1
    amb = alpha - beta
    if amb != 1:
        err_msg = (
            err_head
            + "Condition alpha - beta = 1 not fulfilled. "
            + f"alpha={alpha} ; beta={beta} -> alpha - beta = {amb}"
        )
        raise ValueError(err_msg)

    # return benign values for alpha and beta
    return alpha, beta



"""Low-level operations on Keras graph."""
from __future__ import annotations

import inspect
import warnings
from abc import ABCMeta, abstractmethod
from typing import Any, Callable

import numpy as np
import tensorflow.keras.backend as kbackend
import tensorflow.keras.layers as klayers
import tensorflow.keras.models as kmodels

import innvestigate.backend as ibackend
import innvestigate.backend.checks as ichecks
import innvestigate.layers as ilayers
from innvestigate.backend.types import (
    Layer,
    LayerCheck,
    Model,
    NodeDict,
    OptionalList,
    ReverseTensorDict,
    Tensor,
)

__all__ = [
    "get_kernel",
    "get_layer_inbound_count",
    "get_layer_neuronwise_io",
    "copy_layer_wo_activation",
    "copy_layer",
    "pre_softmax_tensors",
    "model_wo_softmax",
    "get_model_layers",
    "model_contains",
    "trace_model_execution",
    "get_model_execution_trace",
    "get_model_execution_graph",
    "print_model_execution_graph",
    "get_bottleneck_nodes",
    "get_bottleneck_tensors",
    "ReverseMappingBase",
    "reverse_model",
]


def get_kernel(layer: Layer) -> np.ndarray:
    """Returns the kernel weights of a layer, i.e, w/o biases."""
    ret = [x for x in layer.get_weights() if len(x.shape) > 1]
    assert len(ret) == 1
    return ret[0]


###############################################################################


def get_layer_inbound_count(layer: Layer) -> int:
    """Returns the number inbound nodes of a layer."""
    return len(layer._inbound_nodes)


def get_layer_neuronwise_io(
    layer: Layer,
    node_index: int = 0,
    Xs: list[Tensor] = None,
    Ys: list[Tensor] = None,
    return_i: bool = True,
    return_o: bool = True,
) -> tuple[list[Tensor], list[Tensor]] | list[Tensor]:
    """Returns the input and output for each neuron in a layer

    Returns the symbolic input and output for each neuron in a layer.
    For a dense layer this is the input output itself.
    For convolutional layers this method extracts for each neuron
    the input output mapping.

    At the moment this function is designed
    to work with dense and conv2d layers.

    :param layer: The targeted layer.
    :param node_index: Index of the layer node to use.
    :param Xs: Ignore the layer's input but use Xs instead.
    :param Ys: Ignore the layer's output but use Ys instead.
    :param return_i: Return the inputs.
    :param return_o: Return the outputs.
    :return: Inputs and outputs, if specified, for each individual neuron.
    """
    if not ichecks.contains_kernel(layer):
        raise NotImplementedError()

    if Xs is None:
        Xs = ibackend.to_list(layer.get_input_at(node_index))
    if Ys is None:
        Ys = ibackend.to_list(layer.get_output_at(node_index))

    if isinstance(layer, klayers.Dense):
        # Xs and Ys are already in shape.
        ret_Xs = Xs
        ret_Ys = Ys
    elif isinstance(layer, klayers.Conv2D):
        kernel = get_kernel(layer)
        # Expect filter dimension to be last.
        n_channels = kernel.shape[-1]

        if return_i:
            extract_patches = ilayers.ExtractConv2DPatches(
                kernel.shape[:2],
                kernel.shape[2],
                layer.strides,
                layer.dilation_rate,
                layer.padding,
            )
            # shape [samples, out_row, out_col, weight_size]
            reshape = ilayers.Reshape((-1, np.product(kernel.shape[:3])))
            ret_Xs = [reshape(extract_patches(x)) for x in Xs]

        if return_o:
            # Get Ys into shape (samples, channels)
            if kbackend.image_data_format() == "channels_first":
                # Ys shape is [samples, channels, out_row, out_col]
                def _reshape(x):
                    x = kbackend.permute_dimensions(x, (0, 2, 3, 1))
                    x = ilayers.Reshape((-1, n_channels))(x)
                    return x

            else:
                # Ys shape is [samples, out_row, out_col, channels]
                def _reshape(x):
                    x = ilayers.Reshape((-1, n_channels))(x)
                    return x

            ret_Ys = [_reshape(x) for x in Ys]

    else:
        raise NotImplementedError()

    # Xs is (n, d) and Ys is (d, channels)
    if return_i and return_o:
        return ret_Xs, ret_Ys
    if return_i:
        return ret_Xs
    if return_o:
        return ret_Ys
    raise Exception()


def get_symbolic_weight_names(layer: Layer, weights: list[Tensor] = None) -> list[str]:
    """Attribute names for weights

    Looks up the attribute names of weight tensors.

    :param layer: Targeted layer.
    :param weights: A list of weight tensors.
    :return: The attribute names of the weights.
    """

    if weights is None:
        weights = layer.weights

    good_guesses = [
        "kernel",
        "bias",
        "gamma",
        "beta",
        "moving_mean",
        "moving_variance",
        "depthwise_kernel",
        "pointwise_kernel",
    ]

    ret = []
    for weight in weights:
        for attr_name in good_guesses + dir(layer):
            if hasattr(layer, attr_name) and id(weight) == id(
                getattr(layer, attr_name)
            ):
                ret.append(attr_name)
                break
    if len(weights) != len(ret):
        raise Exception("Could not find symoblic weight name(s).")

    return ret


def update_symbolic_weights(layer: Layer, weight_mapping: dict[str, Tensor]) -> None:
    """Updates the symbolic tensors of a layer

    Updates the symbolic tensors of a layer by replacing them.

    Note this does not update the loss or anything alike!
    Use with caution!

    :param layer: Targeted layer.
    :param weight_mapping: Dict with attribute name and weight tensors
      as keys and values.
    """

    trainable_weight_ids = [id(x) for x in layer._trainable_weights]
    non_trainable_weight_ids = [id(x) for x in layer._non_trainable_weights]

    for name, weight in weight_mapping.items():
        current_weight = getattr(layer, name)
        current_weight_id = id(current_weight)

        if current_weight_id in trainable_weight_ids:
            idx = trainable_weight_ids.index(current_weight_id)
            layer._trainable_weights[idx] = weight
        else:
            idx = non_trainable_weight_ids.index(current_weight_id)
            layer._non_trainable_weights[idx] = weight

        setattr(layer, name, weight)


def get_layer_from_config(
    old_layer: Layer,
    new_config: dict[str, Any],
    weights: list[np.ndarray] | list[Tensor] | None = None,
    reuse_symbolic_tensors: bool = True,
) -> Layer:
    """Creates a new layer from a config

    Creates a new layer given a changed config and weights etc.

    :param old_layer: A layer that shall be used as base.
    :param new_config: The config to create the new layer.
    :param weights: Weights to set in the new layer.
      Options: np tensors, symbolic tensors, or None,
      in which case the weights from old_layers are used.
    :param reuse_symbolic_tensors: If the weights of the
      old_layer are used copy the symbolic ones or copy
      the Numpy weights.
    :return: The new layer instance.
    """
    new_layer = old_layer.__class__.from_config(new_config)

    if weights is None:
        if reuse_symbolic_tensors:
            weights = old_layer.weights
        else:
            weights = old_layer.get_weights()

    if len(weights) > 0:
        input_shapes = old_layer.get_input_shape_at(0)
        # todo: inspect and set initializers to something fast for speedup
        new_layer.build(input_shapes)

        is_np_weight = [isinstance(x, np.ndarray) for x in weights]
        if all(is_np_weight):
            new_layer.set_weights(weights)
        else:
            if any(is_np_weight):
                raise ValueError(
                    "Expect either all weights to be " "np tensors or symbolic tensors."
                )

            symbolic_names = get_symbolic_weight_names(old_layer)
            update = dict(zip(symbolic_names, weights))
            update_symbolic_weights(new_layer, update)

    return new_layer


def copy_layer_wo_activation(
    layer: Layer,
    keep_bias: bool = True,
    name_template: str | None = None,
    weights: list[np.ndarray] | list[Tensor] | None = None,
    reuse_symbolic_tensors: bool = True,
    **kwargs,
) -> Layer:
    """Copy a Keras layer and remove the activations

    Copies a Keras layer but remove potential activations.

    :param layer: A layer that should be copied.
    :param keep_bias: Keep a potential bias.
    :param weights: Weights to set in the new layer.
      Options: np tensors, symbolic tensors, or None,
      in which case the weights from old_layers are used.
    :param reuse_symbolic_tensors: If the weights of the
      old_layer are used copy the symbolic ones or copy
      the Numpy weights.
    :return: The new layer instance.
    """
    config = layer.get_config()
    if name_template is None:
        config["name"] = None
    else:
        config["name"] = name_template % config["name"]
    if ichecks.contains_activation(layer):
        config["activation"] = None
    if hasattr(layer, "use_bias"):
        if keep_bias is False and config.get("use_bias", True):
            config["use_bias"] = False
            if weights is None:
                if reuse_symbolic_tensors:
                    weights = layer.weights[:-1]
                else:
                    weights = layer.get_weights()[:-1]
    return get_layer_from_config(layer, config, weights=weights, **kwargs)


def copy_layer(
    layer: Layer,
    keep_bias: bool = True,
    name_template: bool = None,
    weights: list[Tensor] | list[np.ndarray] | None = None,
    reuse_symbolic_tensors: bool = True,
    **kwargs,
) -> Layer:
    """Copy a Keras layer.

    :param layer: A layer that should be copied.
    :param keep_bias: Keep a potential bias.
    :param weights: Weights to set in the new layer.
      Options: np tensors, symbolic tensors, or None,
      in which case the weights from old_layers are used.
    :param reuse_symbolic_tensors: If the weights of the
      old_layer are used copy the symbolic ones or copy
      the Numpy weights.
    :return: The new layer instance.
    """
    config = layer.get_config()
    if name_template is None:
        config["name"] = None
    else:
        config["name"] = name_template % config["name"]
    if hasattr(layer, "use_bias"):
        if keep_bias is False and config.get("use_bias", True):
            config["use_bias"] = False
            if weights is None:
                if reuse_symbolic_tensors:
                    weights = layer.weights[:-1]
                else:
                    weights = layer.get_weights()[:-1]
    return get_layer_from_config(layer, config, weights=weights, **kwargs)


def pre_softmax_tensors(Xs: Tensor, should_find_softmax: bool = True) -> list[Tensor]:
    """Finds the tensors that were preceeding a potential softmax."""
    softmax_found = False

    Xs = ibackend.to_list(Xs)
    ret = []
    for x in Xs:
        layer, node_index, _tensor_index = x._keras_history
        if ichecks.contains_activation(layer, activation="softmax"):
            softmax_found = True
            if isinstance(layer, klayers.Activation):
                ret.append(layer.get_input_at(node_index))
            else:
                layer_wo_act = copy_layer_wo_activation(layer)
                ret.append(layer_wo_act(layer.get_input_at(node_index)))

    if should_find_softmax and not softmax_found:
        raise Exception("No softmax found.")

    return ret


def model_wo_softmax(model: Model) -> Model:
    """Creates a new model w/o the final softmax activation."""
    return kmodels.Model(
        inputs=model.inputs, outputs=pre_softmax_tensors(model.outputs), name=model.name
    )


###############################################################################


def get_model_layers(model: Model) -> list[Layer]:
    """Returns all layers of a model."""
    layers = []

    def collect_layers(container: Model) -> None:
        for layer in container.layers:
            if layer in layers:
                raise ValueError(f"Collected layer {layer} twice.")
            layers.append(layer)
            if ichecks.is_module(layer):
                collect_layers(layer)

    collect_layers(model)

    return layers


def model_contains(
    model: Model,
    layer_condition: OptionalList[LayerCheck],
) -> list[list[Layer]]:
    """
    Collect layers in model which satisfy `layer_condition`.
    If multiple conditions are given in `layer_condition`,
    the collected layers are returned for each condition.

    :param model: A Keras model.
    :type model: Model
    :param layer_condition: A boolean function or list of functions that
        check Keras layers.
    :type layer_condition: Union[LayerCheck, List[LayerCheck]]
    :return: List, which for each condition in layer_condition
        contains a list of layers which satisfy that condition.
    :rtype: List[List[Layer]]
    """
    conditions = ibackend.to_list(layer_condition)
    layers = get_model_layers(model)

    # return layers for which condition c holds true
    return [[l for l in layers if c(l)] for c in conditions]


###############################################################################


def apply_mapping_to_fused_bn_layer(mapping, fuse_mode: str = "one_linear") -> Callable:
    """
    Applies a mapping to a linearized Batch Normalization layer.

    :param mapping: The mapping to be applied.
      Should take parameters layer and reverse_state and
      return a mapping function.
    :param fuse_mode: Either 'one_linear': apply the mapping
      to a once linearized layer, or
      'two_linear': apply to twice to a twice linearized layer.
    """
    if fuse_mode not in ["one_linear", "two_linear"]:
        raise ValueError("fuse_mode can only be 'one_linear' or 'two_linear'")

    # TODO (alber): remove this workaround and make a proper class
    def get_scale_layer(kernel, bias):
        _kernel = kernel
        _bias = bias

        class ScaleLayer(klayers.Layer):
            def __init__(self, use_bias=True, **kwargs):
                self._kernel_to_be = _kernel
                self._bias_to_be = _bias
                self.use_bias = use_bias
                super().__init__(**kwargs)

            def get_config(self):
                config = super().get_config()
                config["use_bias"] = self.use_bias
                return config

            def build(self, input_shape):
                def kernel_initializer(_shape, dtype=None):
                    if dtype is not None:
                        warnings.warn(f"Ignore dtype {dtype} as bias type.")
                    return self._kernel_to_be

                self.kernel = self.add_weight(
                    name="kernel",
                    shape=kbackend.int_shape(self._kernel_to_be),
                    initializer=kernel_initializer,
                    trainable=False,
                )
                if self.use_bias:

                    def bias_initializer(_shape, dtype=None):
                        if dtype is not None:
                            warnings.warn(f"Ignore dtype {dtype} as bias type.")
                        return self._bias_to_be

                    self.bias = self.add_weight(
                        name="bias",
                        shape=kbackend.int_shape(self._bias_to_be),
                        initializer=bias_initializer,
                        trainable=False,
                    )
                super().build(input_shape)

            def call(self, inputs, *_args, **_kwargs):
                ret = inputs * self.kernel
                if self.use_bias:
                    ret += self.bias
                return ret

        return ScaleLayer()

    def meta_mapping(layer: Layer, reverse_state: dict):
        # get bn params
        weights = layer.weights.copy()
        if layer.scale:
            gamma = weights.pop(0)
        else:
            gamma = kbackend.ones_like(weights[0])
        if layer.center:
            beta = weights.pop(0)
        else:
            beta = kbackend.zeros_like(weights[0])
        mean, variance = weights

        if fuse_mode == "one_linear":
            tmp = kbackend.sqrt(variance**2 + layer.epsilon)
            tmp_k = gamma / tmp
            tmp_b = -mean / tmp + beta

            inputs = layer.get_input_at(0)
            surrogate_layer = get_scale_layer(tmp_k, tmp_b)
            # init layer
            surrogate_layer(inputs)
            actual_mapping = mapping(surrogate_layer, reverse_state).apply
        else:
            tmp = kbackend.sqrt(variance**2 + layer.epsilon)
            tmp_k1 = 1 / tmp
            tmp_b1 = -mean / tmp
            tmp_k2 = gamma
            tmp_b2 = beta

            inputs = layer.get_input_at(0)
            surrogate_layer1 = get_scale_layer(tmp_k1, tmp_b1)
            surrogate_layer2 = get_scale_layer(tmp_k2, tmp_b2)
            # init layers
            surrogate_layer1(inputs)
            surrogate_layer2(inputs)
            # TODO (alber): update reverse state
            actual_mapping_1 = mapping(surrogate_layer1, reverse_state).apply
            actual_mapping_2 = mapping(surrogate_layer2, reverse_state).apply

            def actual_mapping(
                Xs: list[Tensor],
                Ys: list[Tensor],
                reversed_Ys: list[Tensor],
                reverse_state,
            ):

                X2s = ibackend.apply(surrogate_layer1, Xs)
                # Apply first mapping
                # TODO (alber): update reverse state
                reversed_X2s = actual_mapping_2(X2s, Ys, reversed_Ys, reverse_state)
                return actual_mapping_1(Xs, X2s, reversed_X2s, reverse_state)

        return actual_mapping

    return meta_mapping


###############################################################################


def trace_model_execution(
    model: Model, reapply_on_copied_layers: bool = False
) -> tuple[list[Layer], list[tuple[Layer, list[Tensor], list[Tensor]]], list[Tensor]]:
    """
    Trace and linearize excecution of a model and it's possible containers.
    Return a triple with all layers, a list with a linearized execution
    with (layer, input_tensors, output_tensors), and, possible regenerated,
    outputs of the exectution.

    :param model: A kera model.
    :param reapply_on_copied_layers: If the execution needs to be linearized,
      reapply with copied layers. Might be slow. Prevents changes of the
      original layer's node lists.
    """

    # Get all layers in model.
    layers: list[Layer] = get_model_layers(model)

    # Check if some layers are containers.
    # Ignoring the outermost container, i.e. the passed model.
    contains_container: bool = any(
        l is not model and ichecks.is_module(l) for l in layers
    )

    outputs: list[Tensor]

    # If so rebuild the graph, otherwise recycle computations,
    # and create executed node list. (Keep track of paths?)
    if contains_container is True:
        # When containers/models are used as layers, then layers
        # inside the container/model do not keep track of nodes.
        # This makes it impossible to iterate of the nodes list and
        # recover the input output tensors. (see else clause)
        #
        # To recover the computational graph we need to re-apply it.
        # This implies that the tensors-object we use for the forward
        # pass are different to the passed model. This it not the case
        # for the else clause.
        #
        # Note that reapplying the model does only change the inbound
        # and outbound nodes of the model itself. We copy the model
        # so the passed model should not be affected from the
        # reapplication.

        # For each layer in the model (in forward order),
        # this list tracks the tensors going in and out of it
        executed_nodes: list[tuple[Layer, list[Tensor], list[Tensor]]] = []

        # Monkeypatch the call function in all the used layer classes.
        monkey_patches: list[tuple[Layer, Callable]] = [
            (layer, layer.call) for layer in layers
        ]
        try:

            def patch(self, method: Callable):
                if hasattr(method, "__patched__") is True:
                    raise Exception(
                        "Should not happen as we patch objects, not classes."
                    )

                def patched_fn(*args, **kwargs):
                    input_tensors = args[0]
                    output_tensors = method(*args, **kwargs)
                    executed_nodes.append((self, input_tensors, output_tensors))
                    return output_tensors

                patched_fn.__patched__ = True  # type: ignore
                return patched_fn

            # Apply the patches.
            for layer in layers:
                layer.call = patch(layer, layer.call)

            # Trigger reapplication of model.
            model_copy: Model = kmodels.Model(
                inputs=model.inputs, outputs=model.outputs
            )
            outputs = ibackend.to_list(model_copy(model.inputs))
        finally:
            # Revert the monkey patches
            for layer, old_method in monkey_patches:
                layer.call = old_method

        # Now we have the problem that all the tensors
        # do not have a keras_history attribute as they are not part
        # of any node. Apply the flat model to get it.

        tensor_mapping: dict[Tensor, Tensor] = {tmp: tmp for tmp in model.inputs}
        layer_mapping: dict[Layer, Layer]
        new_executed_nodes: list[tuple[Layer, list[Tensor], list[Tensor]]] = []

        if reapply_on_copied_layers is True:
            layer_mapping = {layer: copy_layer(layer) for layer in layers}
        else:
            layer_mapping = {layer: layer for layer in layers}

        for layer, Xs, Ys in executed_nodes:
            layer = layer_mapping[layer]
            Xs, Ys = ibackend.to_list(Xs), ibackend.to_list(Ys)

            if isinstance(layer, klayers.InputLayer):
                # Special case. Do nothing.
                new_Xs, new_Ys = Xs, Ys
            else:
                new_Xs = [tensor_mapping[x] for x in Xs]
                new_Ys = ibackend.to_list(ibackend.apply(layer, new_Xs))

            # Update values of Ys in tensor_mapping with new_Ys
            tensor_mapping.update(dict(zip(Ys, new_Ys)))
            new_executed_nodes.append((layer, new_Xs, new_Ys))

        layers = [layer_mapping[layer] for layer in layers]
        outputs = [tensor_mapping[x] for x in outputs]
        executed_nodes = new_executed_nodes
    else:
        # Easy and safe way.
        # For each layer in the model (in reverse order),
        # this list tracks the tensors going in and out of it
        reverse_executed_nodes: list[tuple[Layer, list[Tensor], list[Tensor]]]
        reverse_executed_nodes = [
            (
                node.outbound_layer,
                ibackend.to_list(node.input_tensors),
                ibackend.to_list(node.output_tensors),
            )
            for depth in sorted(model._nodes_by_depth.keys())
            for node in model._nodes_by_depth[depth]
        ]
        outputs = model.outputs

        executed_nodes = list(reversed(reverse_executed_nodes))

    # `executed_nodes` potentially contains nodes that are not part
    # of the final execution graph.
    # E.g. if a layer was also applied outside of the model. Then its
    # node list contains nodes that do not contribute to the model's output.
    # Those nodes are filtered here.
    used_as_input = [id(out) for out in outputs]
    tmp = []
    for l, Xs, Ys in reversed(executed_nodes):
        if all(id(Y) in used_as_input for Y in Ys):
            used_as_input += [id(X) for X in Xs]
            tmp.append((l, Xs, Ys))
    executed_nodes = list(reversed(tmp))

    return layers, executed_nodes, outputs


def get_model_execution_trace(
    model: Model,
    keep_input_layers: bool = False,
    reapply_on_copied_layers: bool = False,
) -> list[NodeDict]:
    """
    Returns a list representing the execution graph.
    Each key is the node's id as it is used by the reverse_model method.

    Each associated value contains a dictionary with the following items:

    * `nid`: the node id.
    * `layer`: the layer creating this node.
    * `Xs`: the input tensors (only valid if not in a nested container).
    * `Ys`: the output tensors (only valid if not in a nested container).
    * `Xs_nids`: the ids of the nodes creating the Xs.
    * `Ys_nids`: the ids of nodes using the according output tensor.
    * `Xs_layers`: the layer that created the according input tensor.
    * `Ys_layers`: the layers using the according output tensor.

    :param model: A kera model.
    :param keep_input_layers: Keep input layers.
    :param reapply_on_copied_layers: If the execution needs to be linearized,
      reapply with copied layers. Might be slow. Prevents changes of the
      original layer's node lists.
    """
    # Get execution_trace: list with a linearized execution
    # consisting of `(layer, input_tensors, output_tensors)`.
    execution_trace: list[tuple[Layer, list[Tensor], list[Tensor]]]
    _, execution_trace, _ = trace_model_execution(
        model, reapply_on_copied_layers=reapply_on_copied_layers
    )

    # Enrich execution_trace with node ids to get id_execution_trace
    nid: int | None
    current_nid: int
    id_execution_trace: list[tuple[int | None, Layer, list[Tensor], list[Tensor]]]

    current_nid = 0
    id_execution_trace = []
    for l, Xs, Ys in execution_trace:
        if isinstance(l, klayers.InputLayer):
            id_execution_trace.append((None, l, Xs, Ys))
        else:
            id_execution_trace.append((current_nid, l, Xs, Ys))
            current_nid += 1

    # Create lookups from tensor to creating or receiving layer-node
    inputs_to_node: dict[int, list[int]]
    outputs_to_node: dict[int, int | None]

    inputs_to_node = {}
    outputs_to_node = {}
    for nid, _l, Xs, Ys in id_execution_trace:
        if nid is not None:
            for X in Xs:
                Xid = id(X)
                if Xid in inputs_to_node:
                    inputs_to_node[Xid].append(nid)
                else:
                    inputs_to_node[Xid] = [nid]

        if keep_input_layers or nid is not None:
            for Y in Ys:
                Yid = id(Y)
                if Yid in inputs_to_node:
                    raise Exception("Cannot be more than one creating node.")
                outputs_to_node[Yid] = nid

    # Enrich trace with this info.
    nid_to_nodes: dict[Layer, tuple[int | None, Layer, list[Tensor], list[Tensor]]]
    model_execution_trace: list[NodeDict]

    # TODO: fix invariance of type hints
    Xs_nids: list[int | None]
    Ys_nids: list[list[int] | list[None]]
    # Xs_layers = List[Layer]
    # Ys_layers = List[List[Layer]]

    nid_to_nodes = {t[0]: t for t in id_execution_trace}

    model_execution_trace = []
    for nid, l, Xs, Ys in id_execution_trace:

        if isinstance(l, klayers.InputLayer):
            # The nids that created or receive the tensors.
            Xs_nids = []  # Input layer does not receive.
            Ys_nids = [inputs_to_node[id(Y)] for Y in Ys]
            # The layers that created or receive the tensors.
            Xs_layers = []  # Input layer does not receive.
            Ys_layers = [
                [nid_to_nodes[Ynid][1] for Ynid in Ynids2] for Ynids2 in Ys_nids
            ]
        else:
            # The nids that created or receive the tensors.
            Xs_nids = [outputs_to_node.get(id(X), None) for X in Xs]
            Ys_nids = [inputs_to_node.get(id(Y), [None]) for Y in Ys]
            # The layers that created or receive the tensors.
            Xs_layers = [nid_to_nodes[Xnid][1] for Xnid in Xs_nids if Xnid is not None]
            Ys_layers = [
                [nid_to_nodes[Ynid][1] for Ynid in Ynids2 if Ynid is not None]
                for Ynids2 in Ys_nids
            ]

        entry: NodeDict = {
            "nid": nid,
            "layer": l,
            "Xs": Xs,
            "Ys": Ys,
            "Xs_nids": Xs_nids,
            "Ys_nids": Ys_nids,
            "Xs_layers": Xs_layers,
            "Ys_layers": Ys_layers,
        }
        model_execution_trace.append(entry)

    if not keep_input_layers:
        model_execution_trace = [
            t for t in model_execution_trace if t["nid"] is not None
        ]

    return model_execution_trace


def get_model_execution_graph(
    model: Model, keep_input_layers: bool = False
) -> dict[int | None, NodeDict | list[NodeDict]]:
    """
    Returns a dictionary representing the execution graph.
    Each key is the node's id as it is used by the reverse_model method.

    Each associated value contains a dictionary with the following items:

    * `nid`: the node id.
    * `layer`: the layer creating this node.
    * `Xs`: the input tensors (only valid if not in a nested container).
    * `Ys`: the output tensors (only valid if not in a nested container).
    * `Xs_nids`: the ids of the nodes creating the Xs.
    * `Ys_nids`: the ids of nodes using the according output tensor.
    * `Xs_layers`: the layer that created the according input tensor.
    * `Ys_layers`: the layers using the according output tensor.

    :param model: A kera model.
    :param keep_input_layers: Keep input layers.
    """
    trace: list[NodeDict]
    input_layers: list[NodeDict]
    graph: dict[int | None, NodeDict | list[NodeDict]]

    trace = get_model_execution_trace(
        model, keep_input_layers=keep_input_layers, reapply_on_copied_layers=False
    )

    # Input layers in graph have Node-ID `nid=None`.
    # Extract these from list of all nodes `trace`.
    input_layers = [node for node in trace if node["nid"] is None]

    # Create graph which maps Node-IDs to node information from full trace.
    graph = {node["nid"]: node for node in trace}
    if keep_input_layers:
        graph[None] = input_layers

    return graph


def print_model_execution_graph(
    graph: dict[int | None, OptionalList[NodeDict]]
) -> None:
    """Pretty print of a model execution graph."""
    # TODO: check types

    def nids_as_str(nids: list[int | None]) -> str:  # type: ignore
        return ", ".join([str(nid) for nid in nids])  # type: ignore

    def print_node(node) -> None:
        print(
            f"""[NID: {node["nid"]:d}]
            [Layer: {node["layer"].name:20s}]
            [Inputs from: {nids_as_str(node["Xs_nids"]):20s}]
            [Outputs to: {nids_as_str(node["Ys_nids"]):20s}]"""
        )

    def print_input_node(node) -> None:  # node of type NodeDict?
        print(
            f"""[Layer: {node["layer"].name:20s}]
            [Outputs to: {nids_as_str(node["Ys_nids"]):20s}]"""
        )

    if None in graph:  # Input layers in graph have Node-ID `None`
        print("Graph input layers:")
        for input_node in graph[None]:
            print_input_node(input_node)

    print("Graph nodes:")
    for nid in sorted(key for key in graph if key is not None):
        if nid is None:
            continue
        print_node(graph[nid])


def get_bottleneck_nodes(
    inputs: list[Tensor],
    outputs: list[Tensor],
    execution_list: list[tuple[Layer, list[Tensor], list[Tensor]]],
) -> list[tuple[Layer, tuple[list[Tensor], list[Tensor]]]]:
    """
    Given an execution list this function returns all nodes that
    are a bottleneck in the network, i.e., "all information" must pass
    through this node.
    """
    forward_connections: dict[int, list[Tensor]] = {}
    for l, Xs, Ys in execution_list:
        if isinstance(l, klayers.InputLayer):
            # Special case, do nothing.
            continue

        for X in Xs:
            if id(X) in forward_connections:
                forward_connections[id(X)] += Ys
            else:
                forward_connections[id(X)] = list(Ys)

    open_connections: dict[int, bool] = {}
    for X in inputs:
        for fw_c in forward_connections[id(X)]:
            open_connections[id(fw_c)] = True

    ret = []
    for l, Xs, Ys in execution_list:
        if isinstance(l, klayers.InputLayer):
            # Special case, do nothing.
            # Note: if a single input branches
            # this is not detected.
            continue

        for Y in Ys:
            assert id(Y) in open_connections
            del open_connections[id(Y)]

        if len(open_connections) == 0:
            ret.append((l, (Xs, Ys)))

        for Y in Ys:
            if Y not in outputs:
                for fwc in forward_connections[id(Y)]:
                    open_connections[id(fwc)] = True

    return ret


def get_bottleneck_tensors(
    inputs: list[Tensor],
    outputs: list[Tensor],
    execution_list: list[tuple[Layer, list[Tensor], list[Tensor]]],
) -> list[Tensor]:
    """
    Given an execution list this function returns all tensors that
    are a bottleneck in the network, i.e., "all information" must pass
    through this tensor.
    """
    nodes: list[tuple[Layer, tuple[list[Tensor], list[Tensor]]]]
    nodes = get_bottleneck_nodes(inputs, outputs, execution_list)

    ret: dict[int, Tensor] = {}
    for _l, (Xs, Ys) in nodes:
        for tensor_list in (Xs, Ys):
            if len(tensor_list) == 1:
                tensor = tensor_list[0]
                if id(tensor) not in ret:
                    ret[id(tensor)] = tensor
            else:
                # TODO(albermax): put warning here?
                pass
    return list(ret.values())


###############################################################################


class ReverseMappingBase(metaclass=ABCMeta):
    @abstractmethod
    def apply(
        self,
        Xs: list[Tensor],
        Ys: list[Tensor],
        Rs: list[Tensor],
        reverse_state: dict,
    ) -> list[Tensor]:
        pass


def reverse_model(
    model: Model,
    reverse_mappings,  # TODO: type annotate reverse_mappings
    default_reverse_mapping: Callable | None = None,
    head_mapping: Callable = None,
    stop_mapping_at_tensors: list[Tensor] = None,
    verbose: bool = False,
    return_all_reversed_tensors: bool = False,
    clip_all_reversed_tensors: bool | tuple[float, float] = False,
    project_bottleneck_tensors: bool | tuple[float, float] = False,
    execution_trace: None
    | (
        tuple[list[Layer], list[tuple[Layer, list[Tensor], list[Tensor]]], list[Tensor]]
    ) = None,
    reapply_on_copied_layers: bool = False,
) -> tuple[list[Tensor], dict[Tensor, ReverseTensorDict] | None]:
    """
    Reverses a Keras model based on the given reverse functions.
    Returns two values:
    1. the reverted tensors for the according model inputs.
    2. If `return_all_reversed_tensors` is true, a dictionary of all reversed tensors,
        otherwise None.

    :param model: A Keras model.
    :param reverse_mappings: Either a callable that matches layers to
      mappings or a dictionary with layers as keys and mappings as values.
      Allowed as mapping forms are:
          * A function of form (A) f(Xs, Ys, reversed_Ys, reverse_state).
          * A function of form f(B) f(layer, reverse_state) that returns
            a function of form (A).
          * A :class:`ReverseMappingBase` subclass.
    :param default_reverse_mapping: A function that reverses layers for
      which no mapping was given by param "reverse_mappings".
    :param head_mapping: Map output tensors to new values before passing
      them into the reverted network.
    :param stop_mapping_at_tensors: Tensors at which to stop the mapping.
      Similar to stop_gradient parameters for gradient computation.
    :param verbose: Print what's going on.
    :param return_all_reversed_tensors: Return all reverted tensors in addition
      to reverted model input tensors.
    :param clip_all_reversed_tensors: Clip each reverted tensor. False or tuple
      with min/max value.
    :param project_bottleneck_tensors: Project bottleneck layers in the
      reverting process into a given value range. False, True or (a, b) for
      projection range.
    :param reapply_on_copied_layers: When a model execution needs to
      linearized and copy layers before reapplying them. See
      :func:`trace_model_execution`.
    """

    # Set default values ######################################################
    if stop_mapping_at_tensors is None:
        stop_mapping_at_tensors = []

    stop_mapping_at_ids = [id(X) for X in stop_mapping_at_tensors]

    if head_mapping is None:

        def head_mapping(X):
            return X

    if not callable(reverse_mappings):
        # not callable, assume a dict that maps from layer to mapping
        reverse_mapping_data = reverse_mappings

        def reverse_mappings(layer):
            try:
                return reverse_mapping_data[type(layer)]
            except KeyError:
                return None

    if clip_all_reversed_tensors is True:
        raise NotImplementedError(
            "Keyword argument `clip_all_reversed_tensors` ",
            "expected to be `False` or tuple with min/max values.",
        )

    def _print(x):
        if verbose is True:
            print(x)

    # Initialize structure that keeps track of reversed tensors
    # maps tensor to reverse tensor and additional node information
    reversed_tensors: dict[int, ReverseTensorDict]
    reversed_tensors = {}

    bottleneck_tensor_ids: set[int] = set()

    def add_reversed_tensors(nid, tensors_list, reversed_tensors_list) -> None:
        def add_reversed_tensor(i, X: Tensor, reversed_X: Tensor) -> None:
            # reversed_X corresponds to the reverse-propagated relevance
            # or the output (e.g. max neuron activation)

            # Do not keep tensors that should stop the mapping.
            if id(X) in stop_mapping_at_ids:
                return

            if reversed_X is None:
                raise TypeError(
                    "Propagated relevance `reversed_X` is None, "
                    "is expected to be Tensor."
                )

            if (
                id(X) not in reversed_tensors
            ):  # no duplicate entries for forward tensors
                reversed_tensors[id(X)] = {
                    "nid": (nid, i),
                    "tensors": [reversed_X],
                    "final_tensor": None,
                }
            else:  # more than one tensor propagating relevance to X
                reversed_tensors[id(X)]["tensors"].append(reversed_X)

        tmp = zip(tensors_list, reversed_tensors_list)
        for i, (X, reversed_X) in enumerate(tmp):
            add_reversed_tensor(i, X, reversed_X)

    def get_reversed_tensor(tensor: Tensor) -> Tensor:
        tmp: ReverseTensorDict
        tmp = reversed_tensors[id(tensor)]

        if tmp["final_tensor"] is None:
            if len(tmp["tensors"]) == 1:
                final_tensor = tmp["tensors"][0]
            elif len(tmp["tensors"]) > 1:
                final_tensor = klayers.Add()(tmp["tensors"])
            else:
                raise RuntimeError(
                    f"Error during graph reversal: no tensors connected to {tensor}."
                )

            if project_bottleneck_tensors is True:
                if id(tensor) in bottleneck_tensor_ids:
                    project = ilayers.Project(project_bottleneck_tensors)
                    final_tensor = project(final_tensor)

            if isinstance(clip_all_reversed_tensors, tuple):
                clip = ilayers.Clip(*clip_all_reversed_tensors)
                final_tensor = clip(final_tensor)

            tmp["final_tensor"] = final_tensor

        return tmp["final_tensor"]

    # Reverse the model #######################################################
    _print(f"Reverse model: {model}")

    # Create a list with nodes in reverse execution order.
    if execution_trace is None:
        execution_trace = trace_model_execution(
            model, reapply_on_copied_layers=reapply_on_copied_layers
        )
    layers, execution_list, outputs = execution_trace
    reverse_execution_list = reversed(execution_list)

    len_execution_list = len(execution_list)
    num_input_layers = sum(isinstance(l, klayers.InputLayer) for l in layers)
    len_execution_list_wo_inputs_layers = len_execution_list - num_input_layers

    # Initialize the reverse mapping functions.
    initialized_reverse_mappings: dict[Layer, Callable]  # TODO: specify Callable
    initialized_reverse_mappings = {}
    for layer in layers:
        # A layer can be shared, i.e., applied several times.
        # Allow to share a ReverMappingBase for each layer instance
        # in order to reduce the overhead.

        meta_reverse_mapping = reverse_mappings(layer)
        if meta_reverse_mapping is None:
            reverse_mapping = default_reverse_mapping
        elif inspect.isclass(meta_reverse_mapping) and issubclass(
            meta_reverse_mapping, ReverseMappingBase
        ):
            # Mapping is a class
            reverse_mapping_obj = meta_reverse_mapping(
                layer,
                {
                    "model": model,
                    "layer": layer,
                },
            )
            reverse_mapping = reverse_mapping_obj.apply
        else:

            def parameter_count(func):
                if hasattr(inspect, "signature"):
                    ret = len(inspect.signature(func).parameters)
                else:
                    spec = inspect.getargspec(func)
                    ret = len(spec.args)
                    if spec.varargs is not None:
                        ret += len(spec.varargs)
                    if spec.keywords is not None:
                        ret += len(spec.keywords)
                    if ret == 3:
                        # assume class function with self
                        ret -= 1
                return ret

            if (
                callable(meta_reverse_mapping)
                and parameter_count(meta_reverse_mapping) == 2
            ):
                # Function that returns mapping
                reverse_mapping = meta_reverse_mapping(
                    layer,
                    {
                        "model": model,
                        "layer": layer,
                    },
                )
            else:
                # Nothing meta here
                reverse_mapping = meta_reverse_mapping

        # TODO: add annotations
        initialized_reverse_mappings[layer] = reverse_mapping  # type: ignore

    if project_bottleneck_tensors:
        ids = [
            id(X) for X in get_bottleneck_tensors(model.inputs, outputs, execution_list)
        ]
        bottleneck_tensor_ids.update(ids)

    # Initialize the reverse tensor mappings.
    add_reversed_tensors(-1, outputs, [head_mapping(tmp) for tmp in outputs])  # type: ignore # noqa

    # Follow the list and revert the graph.
    for _nid, (layer, Xs, Ys) in enumerate(reverse_execution_list):
        nid = len_execution_list_wo_inputs_layers - _nid - 1

        if isinstance(layer, klayers.InputLayer):
            # Special case. Do nothing.
            pass
        elif ichecks.is_module(layer):
            raise Exception("This is not supposed to happen!")
        else:
            Xs, Ys = ibackend.to_list(Xs), ibackend.to_list(Ys)
            if not all(id(Y) in reversed_tensors for Y in Ys):
                # This node is not part of our computational graph.
                # The (node-)world is bigger than this model.
                # Potentially this node is also not part of the
                # reversed tensor set because it depends on a tensor
                # that is listed in stop_mapping_at_ids.
                continue
            reversed_Ys = [get_reversed_tensor(Y) for Y in Ys]
            local_stop_mapping_at_ids = [
                id(X) for X in Xs if id(X) in stop_mapping_at_ids
            ]

            _print(f"[NID: {nid}] Reverse layer-node {layer}")
            reverse_mapping = initialized_reverse_mappings[layer]
            reversed_Xs = reverse_mapping(
                Xs,
                Ys,
                reversed_Ys,
                {
                    "nid": nid,
                    "model": model,
                    "layer": layer,
                    "stop_mapping_at_ids": local_stop_mapping_at_ids,
                },
            )
            reversed_Xs = ibackend.to_list(reversed_Xs)
            add_reversed_tensors(nid, Xs, reversed_Xs)

    # Return requested values
    reversed_input_tensors = [
        get_reversed_tensor(tmp)
        for tmp in model.inputs
        if id(tmp) not in stop_mapping_at_ids
    ]
    if return_all_reversed_tensors is True:
        return reversed_input_tensors, reversed_tensors
    return reversed_input_tensors, None



from __future__ import annotations

from innvestigate.analyzer.base import AnalyzerBase
from innvestigate.analyzer.deeptaylor import BoundedDeepTaylor, DeepTaylor
from innvestigate.analyzer.gradient_based import (
    BaselineGradient,
    Deconvnet,
    Gradient,
    GuidedBackprop,
    InputTimesGradient,
    IntegratedGradients,
    SmoothGrad,
)
from innvestigate.analyzer.misc import Input, Random
from innvestigate.analyzer.relevance_based.relevance_analyzer import (
    LRP,
    LRPZ,
    BaselineLRPZ,
    LRPAlpha1Beta0,
    LRPAlpha1Beta0IgnoreBias,
    LRPAlpha2Beta1,
    LRPAlpha2Beta1IgnoreBias,
    LRPAlphaBeta,
    LRPEpsilon,
    LRPFlat,
    LRPSequentialPresetA,
    LRPSequentialPresetAFlat,
    LRPSequentialPresetB,
    LRPSequentialPresetBFlat,
    LRPSequentialPresetBFlatUntilIdx,
    LRPWSquare,
    LRPZPlus,
    LRPZPlusFast,
)
from innvestigate.analyzer.wrapper import AugmentReduceBase  # noqa
from innvestigate.analyzer.wrapper import GaussianSmoother  # noqa
from innvestigate.analyzer.wrapper import PathIntegrator  # noqa
from innvestigate.analyzer.wrapper import WrapperBase  # noqa
from innvestigate.backend.types import Model

# Silence flake8
assert BaselineLRPZ

analyzers: dict[str, type[AnalyzerBase]] = {
    # Utility.
    "input": Input,
    "random": Random,
    # Gradient based
    "gradient": Gradient,
    "gradient.baseline": BaselineGradient,
    "input_t_gradient": InputTimesGradient,
    "deconvnet": Deconvnet,
    "guided_backprop": GuidedBackprop,
    "integrated_gradients": IntegratedGradients,
    "smoothgrad": SmoothGrad,
    # Relevance based
    "lrp": LRP,
    "lrp.z": LRPZ,
    "lrp.epsilon": LRPEpsilon,
    "lrp.w_square": LRPWSquare,
    "lrp.flat": LRPFlat,
    "lrp.alpha_beta": LRPAlphaBeta,
    "lrp.alpha_2_beta_1": LRPAlpha2Beta1,
    "lrp.alpha_2_beta_1_IB": LRPAlpha2Beta1IgnoreBias,
    "lrp.alpha_1_beta_0": LRPAlpha1Beta0,
    "lrp.alpha_1_beta_0_IB": LRPAlpha1Beta0IgnoreBias,
    "lrp.z_plus": LRPZPlus,
    "lrp.z_plus_fast": LRPZPlusFast,
    "lrp.sequential_preset_a": LRPSequentialPresetA,
    "lrp.sequential_preset_b": LRPSequentialPresetB,
    "lrp.sequential_preset_a_flat": LRPSequentialPresetAFlat,
    "lrp.sequential_preset_b_flat": LRPSequentialPresetBFlat,
    "lrp.sequential_preset_b_flat_until_idx": LRPSequentialPresetBFlatUntilIdx,
    # Deep Taylor
    "deep_taylor": DeepTaylor,
    "deep_taylor.bounded": BoundedDeepTaylor,
}


def create_analyzer(name: str, model: Model, **kwargs) -> AnalyzerBase:
    """Instantiates the analyzer with the name 'name'

    This convenience function takes an analyzer name
    creates the respective analyzer.

    Alternatively analyzers can be created directly by
    instantiating the respective classes.

    :param name: Name of the analyzer.
    :param model: The model to analyze, passed to the analyzer's __init__.
    :param kwargs: Additional parameters for the analyzer's .
    :return: An instance of the chosen analyzer.
    :raise KeyError: If there is no analyzer with the passed name.
    """
    try:
        analyzer_class = analyzers[name]
    except KeyError:
        raise KeyError(
            f"No analyzer with the name '{name}' could be found."
            f" All possible names are: {list(analyzers.keys())}"
        )
    return analyzer_class(model, **kwargs)



from __future__ import annotations

import tensorflow.keras.backend as kbackend

from innvestigate.backend.types import Tensor

__all__ = [
    "preprocess_images",
    "postprocess_images",
]


def preprocess_images(images: Tensor, color_coding: str = None) -> Tensor:
    """Image preprocessing

    Takes a batch of images and:
    * Adjust the color axis to the Keras format.
    * Fixes the color coding.

    :param images: Batch of images with 4 axes.
    :param color_coding: Determines the color coding.
      Can be None, 'RGBtoBGR' or 'BGRtoRGB'.
    :return: The preprocessed batch.
    """

    ret: Tensor = images
    image_data_format: str = kbackend.image_data_format()

    # TODO: not very general:
    channels_first: bool = images.shape[1] in [1, 3]
    if image_data_format == "channels_first" and not channels_first:
        ret = ret.transpose(0, 3, 1, 2)
    if image_data_format == "channels_last" and channels_first:
        ret = ret.transpose(0, 2, 3, 1)

    assert color_coding in [None, "RGBtoBGR", "BGRtoRGB"]
    if color_coding in ["RGBtoBGR", "BGRtoRGB"]:
        if image_data_format == "channels_first":
            ret = ret[:, ::-1, :, :]
        if image_data_format == "channels_last":
            ret = ret[:, :, :, ::-1]

    return ret


def postprocess_images(
    images: Tensor, color_coding: str = None, channels_first: bool = None
) -> Tensor:
    """Image postprocessing

    Takes a batch of images and reverts the preprocessing.

    :param images: A batch of images with 4 axes.
    :param color_coding: The initial color coding,
      see :func:`preprocess_images`.
    :param channels_first: The output channel format.
    :return: The postprocessed images.
    """

    ret: Tensor = images
    image_data_format: str = kbackend.image_data_format()

    assert color_coding in [None, "RGBtoBGR", "BGRtoRGB"]
    if color_coding in ["RGBtoBGR", "BGRtoRGB"]:
        if image_data_format == "channels_first":
            ret = ret[:, ::-1, :, :]
        if image_data_format == "channels_last":
            ret = ret[:, :, :, ::-1]

    if image_data_format == "channels_first" and not channels_first:
        ret = ret.transpose(0, 2, 3, 1)
    if image_data_format == "channels_last" and channels_first:
        ret = ret.transpose(0, 3, 1, 2)

    return ret



from __future__ import annotations

from typing import Callable

import numpy as np
import tensorflow.keras.backend as kbackend

import innvestigate.backend as ibackend
import innvestigate.backend.graph as igraph
import innvestigate.layers as ilayers
from innvestigate.analyzer.network_base import AnalyzerNetworkBase
from innvestigate.backend.types import (
    CondReverseMapping,
    Layer,
    Model,
    OptionalList,
    ReverseTensorDict,
    Tensor,
)

__all__ = ["ReverseAnalyzerBase"]


class ReverseAnalyzerBase(AnalyzerNetworkBase):
    """Convenience class for analyzers that revert the model's structure.

    This class contains many helper functions around the graph
    reverse function :func:`innvestigate.backend.graph.reverse_model`.

    The deriving classes should specify how the graph should be reverted
    by implementing the following functions:

    * :func:`_reverse_mapping(layer)` given a layer this function
      returns a reverse mapping for the layer as specified in
      :func:`innvestigate.backend.graph.reverse_model` or None.

      This function can be implemented, but it is encouraged to
      implement a default mapping and add additional changes with
      the function :func:`_add_conditional_reverse_mapping` (see below).

      The default behavior is finding a conditional mapping (see below),
      if none is found, :func:`_default_reverse_mapping` is applied.
    * :func:`_default_reverse_mapping` defines the default
      reverse mapping.
    * :func:`_head_mapping` defines how the outputs of the model
      should be instantiated before the are passed to the reversed
      network.

    Furthermore other parameters of the function
    :func:`innvestigate.backend.graph.reverse_model` can
    be changed by setting the according parameters of the
    init function:

    :param reverse_verbose: Print information on the reverse process.
    :param reverse_clip_values: Clip the values that are passed along
      the reverted network. Expects tuple (min, max).
    :param reverse_project_bottleneck_layers: Project the value range
      of bottleneck tensors in the reverse network into another range.
    :param reverse_check_min_max_values: Print the min/max values
      observed in each tensor along the reverse network whenever
      :func:`analyze` is called.
    :param reverse_check_finite: Check if values passed along the
      reverse network are finite.
    :param reverse_keep_tensors: Keeps the tensors created in the
      backward pass and stores them in the attribute
      :attr:`_reversed_tensors`.
    :param reverse_reapply_on_copied_layers: See
      :func:`innvestigate.backend.graph.reverse_model`.
    """

    def __init__(
        self,
        model: Model,
        reverse_verbose: bool = False,
        reverse_clip_values: bool = False,
        reverse_project_bottleneck_layers: bool = False,
        reverse_check_min_max_values: bool = False,
        reverse_check_finite: bool = False,
        reverse_keep_tensors: bool = False,
        reverse_reapply_on_copied_layers: bool = False,
        **kwargs,
    ) -> None:
        """
        From AnalyzerBase super init:
        * Initializes empty list of _model_checks

        From AnalyzerNetworkBase super init:
        * set _neuron_selection_mode
        * add check for lambda layers through 'allow_lambda_layers'
        * define attributes for '_prepare_model', which is later called
            through 'create_analyzer_model'

        Here:
        * define attributes required for calling '_conditional_reverse_mapping'
        """
        super().__init__(model, **kwargs)

        self._reverse_verbose = reverse_verbose
        self._reverse_clip_values = reverse_clip_values
        self._reverse_project_bottleneck_layers = reverse_project_bottleneck_layers
        self._reverse_check_min_max_values = reverse_check_min_max_values
        self._reverse_check_finite = reverse_check_finite
        self._reverse_keep_tensors = reverse_keep_tensors
        self._reverse_reapply_on_copied_layers = reverse_reapply_on_copied_layers
        self._reverse_mapping_applied: bool = False

        # map priorities to lists of conditional reverse mappings
        self._conditional_reverse_mappings: dict[int, list[CondReverseMapping]] = {}

        # Maps keys "min", "max", "finite", "keep" to tuples of indices
        self._debug_tensors_indices: dict[str, tuple[int, int]] = {}

    def _gradient_reverse_mapping(
        self,
        Xs: OptionalList[Tensor],
        Ys: OptionalList[Tensor],
        reversed_Ys: OptionalList[Tensor],
        reverse_state: dict,
    ):
        """Returns masked gradient."""
        mask = [id(X) not in reverse_state["stop_mapping_at_ids"] for X in Xs]
        grad = ibackend.gradients(Xs, Ys, reversed_Ys)
        return ibackend.apply_mask(grad, mask)

    def _reverse_mapping(self, layer: Layer):
        """
        This function should return a reverse mapping for the passed layer.

        If this function returns None, :func:`_default_reverse_mapping`
        is applied.

        :param layer: The layer for which a mapping should be returned.
        :return: The mapping can be of the following forms:
          * A function of form (A) f(Xs, Ys, reversed_Ys, reverse_state)
            that maps reversed_Ys to reversed_Xs (which should contain
            tensors of the same shape and type).
          * A function of form f(B) f(layer, reverse_state) that returns
            a function of form (A).
          * A :class:`ReverseMappingBase` subclass.
        """
        if layer in self._special_helper_layers:
            # Special layers added by AnalyzerNetworkBase
            # that should not be exposed to user.
            return self._gradient_reverse_mapping

        return self._apply_conditional_reverse_mappings(layer)

    def _add_conditional_reverse_mapping(
        self,
        condition: Callable[[Layer], bool],
        mapping: Callable,  # TODO: specify type of Callable
        priority: int = -1,
        name: str | None = None,
    ):
        """
        This function should return a reverse mapping for the passed layer.

        If this function returns None, :func:`_default_reverse_mapping`
        is applied.

        :param condition: Condition when this mapping should be applied.
          Form: f(layer) -> bool
        :param mapping: The mapping can be of the following forms:
          * A function of form (A) f(Xs, Ys, reversed_Ys, reverse_state)
            that maps reversed_Ys to reversed_Xs (which should contain
            tensors of the same shape and type).
          * A function of form f(B) f(layer, reverse_state) that returns
            a function of form (A).
          * A :class:`ReverseMappingBase` subclass.
        :param priority: The higher the earlier the condition gets
          evaluated.
        :param name: An identifying name.
        """
        if self._reverse_mapping_applied is True:
            raise Exception(
                "Cannot add conditional mapping " "after first application."
            )

        # Add key `priority` to dict _conditional_reverse_mappings
        # if it doesn't exist yet.
        if priority not in self._conditional_reverse_mappings:
            self._conditional_reverse_mappings[priority] = []

        # Add Conditional Reveserse mapping at given priority
        tmp: CondReverseMapping = {
            "condition": condition,
            "mapping": mapping,
            "name": name,
        }
        self._conditional_reverse_mappings[priority].append(tmp)

    def _apply_conditional_reverse_mappings(self, layer):
        mappings = getattr(self, "_conditional_reverse_mappings", {})
        self._reverse_mapping_applied = True

        # Search for mapping. First consider ones with highest priority,
        # inside priority in order of adding.
        sorted_keys = reversed(sorted(mappings.keys()))
        for key in sorted_keys:
            for mapping in mappings[key]:
                if mapping["condition"](layer):
                    return mapping["mapping"]

        # Otherwise return None and default reverse mapping will be applied
        return None

    def _default_reverse_mapping(
        self,
        Xs: OptionalList[Tensor],
        Ys: OptionalList[Tensor],
        reversed_Ys: OptionalList[Tensor],
        reverse_state: dict,
    ):
        """
        Fallback function to map reversed_Ys to reversed_Xs
        (which should contain tensors of the same shape and type).
        """
        return self._gradient_reverse_mapping(Xs, Ys, reversed_Ys, reverse_state)

    def _head_mapping(self, X: Tensor) -> Tensor:
        """
        Map output tensors to new values before passing
        them into the reverted network.
        """
        # Here: Keep the output signal.
        # Should be re-implemented by inheritance.
        # Refer to the "Introduction to development notebook".
        return X

    def _postprocess_analysis(self, Xs: OptionalList[Tensor]) -> list[Tensor]:
        return ibackend.to_list(Xs)

    def _reverse_model(
        self,
        model: Model,
        stop_analysis_at_tensors: list[Tensor] = None,
        return_all_reversed_tensors=False,
    ) -> tuple[list[Tensor], dict[Tensor, ReverseTensorDict] | None]:
        if stop_analysis_at_tensors is None:
            stop_analysis_at_tensors = []

        return igraph.reverse_model(
            model,
            reverse_mappings=self._reverse_mapping,
            default_reverse_mapping=self._default_reverse_mapping,
            head_mapping=self._head_mapping,
            stop_mapping_at_tensors=stop_analysis_at_tensors,
            verbose=self._reverse_verbose,
            clip_all_reversed_tensors=self._reverse_clip_values,
            project_bottleneck_tensors=self._reverse_project_bottleneck_layers,
            return_all_reversed_tensors=return_all_reversed_tensors,
        )

    def _create_analysis(
        self, model: Model, stop_analysis_at_tensors: list[Tensor] = None
    ):

        if stop_analysis_at_tensors is None:
            stop_analysis_at_tensors = []

        return_all_reversed_tensors = (
            self._reverse_check_min_max_values
            or self._reverse_check_finite
            or self._reverse_keep_tensors
        )

        # if return_all_reversed_tensors is False,
        # reversed_tensors will be None
        reversed_input_tensors, reversed_tensors = self._reverse_model(
            model,
            stop_analysis_at_tensors=stop_analysis_at_tensors,
            return_all_reversed_tensors=return_all_reversed_tensors,
        )
        ret = self._postprocess_analysis(reversed_input_tensors)

        if return_all_reversed_tensors:
            if reversed_tensors is None:
                raise TypeError("Expected reversed_tensors, got None.")

            debug_tensors: list[Tensor]
            tmp: list[Tensor]

            debug_tensors = []
            values = reversed_tensors.values()
            mapping = {i: v["nid"] for i, v in enumerate(values)}
            tensors = [v["final_tensor"] for v in values]
            self._reverse_tensors_mapping = mapping

            if self._reverse_check_min_max_values:
                tmp = [kbackend.min(x) for x in tensors]
                self._debug_tensors_indices["min"] = (
                    len(debug_tensors),
                    len(debug_tensors) + len(tmp),
                )
                debug_tensors += tmp

                tmp = [kbackend.min(x) for x in tensors]
                self._debug_tensors_indices["max"] = (
                    len(debug_tensors),
                    len(debug_tensors) + len(tmp),
                )
                debug_tensors += tmp

            if self._reverse_check_finite:
                tmp = ibackend.to_list(ilayers.FiniteCheck()(tensors))
                self._debug_tensors_indices["finite"] = (
                    len(debug_tensors),
                    len(debug_tensors) + len(tmp),
                )
                debug_tensors += tmp

            if self._reverse_keep_tensors:
                self._debug_tensors_indices["keep"] = (
                    len(debug_tensors),
                    len(debug_tensors) + len(tensors),
                )
                debug_tensors += tensors

            return ret, debug_tensors
        return ret

    def _handle_debug_output(self, debug_values):

        if self._reverse_check_min_max_values:
            indices = self._debug_tensors_indices["min"]
            tmp = debug_values[indices[0] : indices[1]]
            tmp = sorted(
                (self._reverse_tensors_mapping[i], v) for i, v in enumerate(tmp)
            )
            print(f"Minimum values in tensors: ((NodeID, TensorID), Value) - {tmp}")

            indices = self._debug_tensors_indices["max"]
            tmp = debug_values[indices[0] : indices[1]]
            tmp = sorted(
                (self._reverse_tensors_mapping[i], v) for i, v in enumerate(tmp)
            )
            print(f"Maximum values in tensors: ((NodeID, TensorID), Value) - {tmp}")

        if self._reverse_check_finite:
            indices = self._debug_tensors_indices["finite"]
            tmp = debug_values[indices[0] : indices[1]]
            nfinite_tensors = np.flatnonzero(np.asarray(tmp) > 0)

            if len(nfinite_tensors) > 0:
                nfinite_tensors = sorted(
                    self._reverse_tensors_mapping[i] for i in nfinite_tensors
                )
                print(
                    "Not finite values found in following nodes: "
                    f"(NodeID, TensorID) - {nfinite_tensors}"
                )

        if self._reverse_keep_tensors:
            indices = self._debug_tensors_indices["keep"]
            tmp = debug_values[indices[0] : indices[1]]
            tmp = sorted(
                (self._reverse_tensors_mapping[i], v) for i, v in enumerate(tmp)
            )
            self._reversed_tensors = tmp

    def _get_state(self):
        state = super()._get_state()
        state.update(
            {
                "reverse_verbose": self._reverse_verbose,
                "reverse_clip_values": self._reverse_clip_values,
                "reverse_project_bottleneck_layers": self._reverse_project_bottleneck_layers,  # noqa
                "reverse_check_min_max_values": self._reverse_check_min_max_values,
                "reverse_check_finite": self._reverse_check_finite,
                "reverse_keep_tensors": self._reverse_keep_tensors,
                "reverse_reapply_on_copied_layers": self._reverse_reapply_on_copied_layers,  # noqa
            }
        )
        return state

    @classmethod
    def _state_to_kwargs(cls, state: dict):
        reverse_verbose = state.pop("reverse_verbose")
        reverse_clip_values = state.pop("reverse_clip_values")
        reverse_project_bottleneck_layers = state.pop(
            "reverse_project_bottleneck_layers"
        )
        reverse_check_min_max_values = state.pop("reverse_check_min_max_values")
        reverse_check_finite = state.pop("reverse_check_finite")
        reverse_keep_tensors = state.pop("reverse_keep_tensors")
        reverse_reapply_on_copied_layers = state.pop("reverse_reapply_on_copied_layers")
        # call super after popping class-specific states:
        kwargs = super()._state_to_kwargs(state)

        kwargs.update(
            {
                "reverse_verbose": reverse_verbose,
                "reverse_clip_values": reverse_clip_values,
                "reverse_project_bottleneck_layers": reverse_project_bottleneck_layers,
                "reverse_check_min_max_values": reverse_check_min_max_values,
                "reverse_check_finite": reverse_check_finite,
                "reverse_keep_tensors": reverse_keep_tensors,
                "reverse_reapply_on_copied_layers": reverse_reapply_on_copied_layers,
            }
        )
        return kwargs



from __future__ import annotations

from typing import Any

import tensorflow.keras.layers as klayers
import tensorflow.keras.models as kmodels

import innvestigate.analyzer.relevance_based.relevance_rule as lrp_rules
import innvestigate.backend.checks as ichecks
import innvestigate.backend.graph as igraph
from innvestigate.analyzer.reverse_base import ReverseAnalyzerBase
from innvestigate.backend.types import Model

__all__ = [
    "DeepTaylor",
    "BoundedDeepTaylor",
]


class DeepTaylor(ReverseAnalyzerBase):
    """DeepTaylor for ReLU-networks with unbounded input

    This class implements the DeepTaylor algorithm for neural networks with
    ReLU activation and unbounded input ranges.

    :param model: A Keras model.
    """

    def __init__(self, model: Model, *args, **kwargs) -> None:
        super().__init__(model, *args, **kwargs)

        # Add and run model checks
        self._add_model_softmax_check()
        self._add_model_check(
            lambda layer: not ichecks.only_relu_activation(layer),
            "This DeepTaylor implementation only supports ReLU activations.",
            check_type="exception",
        )
        self._do_model_checks()

    def _create_analysis(self, *args: Any, **kwargs: Any):

        # Kernel layers.
        self._add_conditional_reverse_mapping(
            lambda l: (ichecks.contains_kernel(l) and ichecks.contains_activation(l)),
            lrp_rules.Alpha1Beta0IgnoreBiasRule,
            name="deep_taylor_kernel_w_relu",
        )
        self._add_conditional_reverse_mapping(
            lambda l: (
                ichecks.contains_kernel(l) and not ichecks.contains_activation(l)
            ),
            lrp_rules.WSquareRule,
            name="deep_taylor_kernel_wo_relu",
        )

        # ReLU Activation layer
        self._add_conditional_reverse_mapping(
            lambda l: (
                not ichecks.contains_kernel(l) and ichecks.only_relu_activation(l)
            ),
            self._gradient_reverse_mapping,
            name="deep_taylor_relu",
        )

        # Assume conv layer beforehand -> unbounded
        bn_mapping = igraph.apply_mapping_to_fused_bn_layer(
            lrp_rules.WSquareRule,
            fuse_mode="one_linear",
        )
        self._add_conditional_reverse_mapping(
            ichecks.is_batch_normalization_layer,
            bn_mapping,
            name="deep_taylor_batch_norm",
        )
        # Special layers.
        self._add_conditional_reverse_mapping(
            ichecks.is_max_pooling,
            self._gradient_reverse_mapping,
            name="deep_taylor_max_pooling",
        )
        self._add_conditional_reverse_mapping(
            ichecks.is_average_pooling,
            self._gradient_reverse_mapping,
            name="deep_taylor_average_pooling",
        )
        self._add_conditional_reverse_mapping(
            lambda l: isinstance(l, klayers.Add),
            # Ignore scaling with 0.5
            self._gradient_reverse_mapping,
            name="deep_taylor_add",
        )
        self._add_conditional_reverse_mapping(
            lambda l: isinstance(
                l,
                (
                    klayers.UpSampling1D,
                    klayers.UpSampling2D,
                    klayers.UpSampling3D,
                    klayers.Dropout,
                    klayers.SpatialDropout1D,
                    klayers.SpatialDropout2D,
                    klayers.SpatialDropout3D,
                ),
            ),
            self._gradient_reverse_mapping,
            name="deep_taylor_special_layers",
        )

        # Layers w/o transformation
        self._add_conditional_reverse_mapping(
            lambda l: isinstance(
                l,
                (
                    klayers.InputLayer,
                    klayers.Cropping1D,
                    klayers.Cropping2D,
                    klayers.Cropping3D,
                    klayers.ZeroPadding1D,
                    klayers.ZeroPadding2D,
                    klayers.ZeroPadding3D,
                    klayers.Concatenate,
                    klayers.Flatten,
                    klayers.Masking,
                    klayers.Permute,
                    klayers.RepeatVector,
                    klayers.Reshape,
                ),
            ),
            self._gradient_reverse_mapping,
            name="deep_taylor_no_transform",
        )

        return super()._create_analysis(*args, **kwargs)

    def _default_reverse_mapping(self, _Xs, _Ys, _reversed_Ys, reverse_state):
        """
        Block all default mappings.
        """
        raise NotImplementedError(f"""Layer {reverse_state["layer"]} not supported.""")

    def _prepare_model(self, model):
        """
        To be theoretically sound Deep-Taylor expects only positive outputs.
        """

        positive_outputs = [klayers.ReLU()(x) for x in model.outputs]
        model_with_positive_output = kmodels.Model(
            inputs=model.inputs, outputs=positive_outputs
        )

        return super()._prepare_model(model_with_positive_output)


class BoundedDeepTaylor(DeepTaylor):
    """DeepTaylor for ReLU-networks with bounded input

    This class implements the DeepTaylor algorithm for neural networks with
    ReLU activation and bounded input ranges.

    :param model: A Keras model.
    :param low: Lowest value of the input range. See Z_B rule.
    :param high: Highest value of the input range. See Z_B rule.
    """

    def __init__(self, model, low=None, high=None, **kwargs):
        super().__init__(model, **kwargs)

        if low is None or high is None:
            raise ValueError(
                "The low or high parameter is missing. "
                "Z-B (bounded rule) require both values."
            )

        self._bounds_low = low
        self._bounds_high = high

    def _create_analysis(self, *args, **kwargs):

        low, high = self._bounds_low, self._bounds_high

        class BoundedProxyRule(lrp_rules.BoundedRule):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, low=low, high=high, **kwargs)

        self._add_conditional_reverse_mapping(
            lambda l: ichecks.is_input_layer(l) and ichecks.contains_kernel(l),
            BoundedProxyRule,
            name="deep_taylor_first_layer_bounded",
            priority=10,  # do first
        )

        return super()._create_analysis(*args, **kwargs)

    def _get_state(self):
        state = super()._get_state()
        state.update({"low": self._bounds_low, "high": self._bounds_high})
        return state

    @classmethod
    def _state_to_kwargs(cls, state):
        low = state.pop("low")
        high = state.pop("high")

        # call super after popping class-specific states:
        kwargs = super()._state_to_kwargs(state)

        kwargs.update({"low": low, "high": high})
        return kwargs



from innvestigate import analyzer  # noqa
from innvestigate.analyzer import create_analyzer  # noqa
from innvestigate.analyzer.base import NotAnalyzeableModelException  # noqa
from innvestigate.backend.graph import model_wo_softmax  # noqa

__version__ = "2.0.0"
